#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""Defines the magic attriobutes that :py:class:`Stoner.Image.ImageFolder` uses."""

__all__ = ["DrawProxy", "MaskProxy"]

from functools import wraps
from io import BytesIO as StreamIO

import numpy as np
from skimage import draw
import matplotlib.pyplot as plt

from ..tools import fix_signature
from ..tools.decorators import class_modifier
from .widgets import ShapeSelect
from .imagefuncs import imshow


def _draw_apaptor(func):
    """Adapt methods for DrawProxy class to bind :py:mod:`skimage.draw` functions."""

    @wraps(func)
    def _proxy(self, *args, **kargs):
        value = kargs.pop("value", np.ones(1, dtype=self._img.dtype)[0])
        coords = func(*args, **kargs)
        if len(coords) == 3:
            rr, cc, vv = coords
            if len(rr) == len(cc):
                coords = rr, cc
                value = value * vv
        if len(coords) == 2 and isinstance(coords[0], np.ndarray) and coords[0].ndim == 3:
            im = type(self._img)(np.zeros(self._img.shape, dtype="uint32"))
            im += coords[0][:, :, 0]
            im += coords[0][:, :, 1] * 256
            im += coords[0][:, :, 2] * 256**2
            im[im == 16777215] = 0
            im.convert(self._img.dtype)
            self._img[im != 0] = im[im != 0]
            return self._parent

        self._img[coords] = value
        return self._parent

    return fix_signature(_proxy, func)


@class_modifier(draw, adaptor=_draw_apaptor, RTD_restrictions=False, no_long_names=True)
class DrawProxy:
    """Provides a wrapper around :py:mod:`skimage.draw` to allow easy drawing of objects onto images.

    This class allows access the user to draw simply shapes on an image (or its mask) by specifying the desired shape
    and geometry (centre, length/width etc). Mostly this implemented by pass throughs to the :py:mod:`skimage.draw`
    module, but methods are provided for an annulus, rectangle (and square) and rectangle-perimeter meothdds- the
    latter offering rotation about the centre point in contrast to the :py:mod:`skimage.draw` equivalents.

    No state data is stored with this class so the attribute does not need to be serialised when the parent ImageFile
    is saved.
    """

    def __init__(self, *args, **kargs):  # pylint: disable=unused-argument
        """Grab the parent image from the constructor."""
        self._img = args[0]
        self._parent = args[1]

    def annulus(self, r, c, radius1, radius2, shape=None, value=1.0):
        """Use a combination of two circles to draw and annulus.

        Args:
            r,c (float): Centre coordinates
            radius1,radius2 (float): Inner and outer radius.

        Keyword Arguments:
            shape (2-tuple, None): Confine the coordinates to staywith shape
            value (float): value to draw with
        Returns:
            A copy of the image with the annulus drawn on it.

        Notes:
            If radius2<radius1 then the sense of the whole shape is inverted
            so that the annulus is left clear and the filed is filled.
        """
        if shape is None:
            shape = self._img.shape
        invert = radius2 < radius1
        if invert:
            buf = np.ones(shape)
            fill = 0.0
            bg = 1.0
        else:
            buf = np.zeros(shape)
            fill = 1.0
            bg = 0.0
        radius1, radius2 = min(radius1, radius2), max(radius1, radius2)
        rr, cc = draw.disk((r, c), radius2, shape=shape)
        buf[rr, cc] = fill
        rr, cc = draw.disk((r, c), radius1, shape=shape)
        buf[rr, cc] = bg
        self._img[:, :] = (self._img * buf + value * (1.0 - buf)).astype(self._img.dtype)
        return self._parent

    if "circle" not in dir(draw):

        def circle(self, r, c, radius, shape=None, value=1.0):
            """ "Generate coordinates of pixels within circle.

            Args:
                r,c (int): coordinates of the centre of the circle to be drawn.
                radius (float): Radius of the circle

            Keyword arguments:
                shape (tuple): Image shape as a tuple of size 2. Determines the maximum extent of output
                    pixel coordinates. This is useful for disks that exceed the image size. If None, the full
                    extent of the disk is used. The shape might result in negative coordinates and wraparound
                    behaviour.
                value (float): pixel value to write with.


            Notes:
                This is actually just a proxy for disk
            """
            return self.disk((r, c), radius, shape=shape, value=value)  # pylint: disable=no-member

    def rectangle(self, r, c, w, h, angle=0.0, shape=None, value=1.0):
        """Draw a rectangle on an image.

        Args:
            r,c (float): Centre coordinates
            w,h (float): Lengths of the two sides of the rectangle

        Keyword Arguments:
            angle (float): Angle to rotate the rectangle about
            shape (2-tuple or None): Confine the coordinates to this shape.
            value (float): The value to draw with.

        Returns:
            A copy of the current image with a rectangle drawn on.
        """
        if shape is None:
            shape = self._img.shape

        x1 = r - h / 2
        x2 = r + h / 2
        y1 = c - w / 2
        y2 = c + w / 2
        co_ords = np.array([[x1, y1], [x2, y1], [x2, y2], [x1, y2]])
        if angle != 0:
            centre = np.array([r, c])
            cos, sin, m = np.cos, np.sin, np.matmul
            r = np.array([[cos(angle), -sin(angle)], [sin(angle), cos(angle)]])
            co_ords = np.array([centre + m(r, xy - centre) for xy in co_ords])
        rr, cc = draw.polygon(co_ords[:, 0], co_ords[:, 1], shape=shape)
        self._img[rr, cc] = value
        return self._parent

    def rectangle_perimeter(self, r, c, w, h, angle=0.0, shape=None, value=1.0):
        """Draw the perimter of a rectangle on an image.

        Args:
            r,c (float): Centre coordinates
            w,h (float): Lengths of the two sides of the rectangle

        Keyword Arguments:
            angle (float): Angle to rotate the rectangle about
            shape (2-tuple or None): Confine the coordinates to this shape.
            value (float): The value to draw with.

        Returns:
            A copy of the current image with a rectangle drawn on.
        """
        if shape is None:
            shape = self._img.shape

        x1 = r - h / 2
        x2 = r + h / 2
        y1 = c - w / 2
        y2 = c + w / 2
        co_ords = np.array([[x1, y1], [x2, y1], [x2, y2], [x1, y2]])
        if angle != 0:
            centre = np.array([r, c])
            cos, sin, m = np.cos, np.sin, np.matmul
            r = np.array([[cos(angle), -sin(angle)], [sin(angle), cos(angle)]])
            co_ords = np.array([centre + m(r, xy - centre) for xy in co_ords])
        rr, cc = draw.polygon_perimeter(co_ords[:, 0], co_ords[:, 1], shape=shape)
        self._img[rr, cc] = value
        return self._parent

    def square(self, r, c, w, angle=0.0, shape=None, value=1.0):
        """Draw a square on an image.

        Args:
            r,c (float): Centre coordinates
            w (float): Length of the side of the square

        Keyword Arguments:
            angle (float): Angle to rotate the rectangle about
            shape (2-tuple or None): Confine the coordinates to this shape.
            value (float): The value to draw with.

        Returns:
            A copy of the current image with a rectangle drawn on.
        """
        return self.rectangle(r, c, w, w, angle=angle, shape=shape, value=value)


class MaskProxy:
    """Provides a wrapper to support manipulating the image mask easily.

    The actual mask of a :py:class:`Stonmer.ImageFile` is held by the mask attribute of the underlying
    :py:class:`numpy.ma.MaskedArray`, but this class implements an attribute for an ImageFile that not only
    provides a means to index into the mask data, but supported other operations too.

    Attributes:
        color (matplotlib colour):
            This defines the colour of the mask that is used when showing a masked image in a window. It can be a
            named colour - such as *red*, *blue* etc. or a tuple of 3 or 4 numbers between 0 and 1 - which define an
            RGB(Alpha) colour. Note that the Alpha channel is actually an opacity channel = so 1.0 is solid and 0 is
            transparent.
        data (numpy array of bool):
            This accesses the actual boolean masked array if it is necessary to get at the full array. It is equivalent
            to .mask[:]
        image (numpoy array of bool):
            This is a synonym for :py:attr:`MaskProxy.data`.
        draw (:py:class:`DrawProxy`):
            This allows the mask to drawn on like the image. This is particularly useful as it provides a convenient
            programmatic way to define regions of interest in the mask with simply geometric shapes.

    Indexing of the MaskProxy simply passes through to the underlying mask data - thus getting, setting and deleting
    element directly changes the mask data.

    The string representation of the mask is an ascii art version of the mask where . is unmasked and X is masked.

    Conversion to a boolean is equaivalent to testing whether **any** elements of the mask are True.

    The mask also supports the invert and negate operators which both return the inverse of the mask (but do not
    change the mask itself - unlike :py:meth:`MaskProxy.invert`).

    For rich displays, the class also supports a png representation which is simply a black and white version of the
    mask with black pixels being masked elements and white unmasked elements.
    """

    @property
    def _imagearray(self):
        """Get the underliying image data."""
        return self._imagefolder.image

    @property
    def _mask(self):
        """Get the mask for the underlying image."""
        self._imagearray.mask = np.ma.getmaskarray(self._imagearray)
        return self._imagearray.mask

    @property
    def colour(self):
        """Get the colour of the mask."""
        return getattr(self._imagearray, "_mask_color", None)

    @colour.setter
    def colour(self, value):
        """Set the colour of the mask."""
        self._imagearray._mask_color = value

    @property
    def data(self):
        """Get the underlying data as an array - compatibility accessor."""
        return self[:]

    @property
    def image(self):
        """Get the underlying data as an array - compatibility accessor."""
        return self[:]

    @property
    def draw(self):
        """Access the draw proxy object."""
        return DrawProxy(self._mask, self._imagefolder)

    def __init__(self, *args):
        """Keep track of the underlying objects."""
        self._imagefolder = args[0]

    def __getitem__(self, index):
        """Proxy through to mask index."""
        return self._mask.__getitem__(index)

    def __setitem__(self, index, value):
        """Proxy through to underlying mask."""
        self._imagearray.mask.__setitem__(index, value)

    def __delitem__(self, index):
        """Proxy through to underlying mask."""
        self._imagearray.mask.__delitem__(index)

    def __getattr__(self, name):
        """Check name against self._imagearray._funcs and constructs a method to edit the mask as an image."""
        if hasattr(self._imagearray.mask, name):
            return getattr(self._imagearray.mask, name)
        func = getattr(type(self._imagearray), name, None)
        if func is None:
            raise AttributeError(f"{name} not a callable mask method.")

        @wraps(func)
        def _proxy_call(*args, **kargs):
            retval = func(self._mask.astype(float).view(type(self._imagearray)) * 1000, *args, **kargs)
            if isinstance(retval, np.ndarray) and retval.shape == self._imagearray.shape:
                retval.normalise()
                self._imagearray.mask = retval > 0
            return retval

        _proxy_call.__doc__ = func.__doc__
        _proxy_call.__name__ = func.__name__
        return fix_signature(_proxy_call, func)

    def __repr__(self):
        """Make a textual representation of the image."""
        output = ""
        f = np.array(["."] * self._mask.shape[1])
        t = np.array(["X"] * self._mask.shape[1])
        for ix in self._mask:
            row = np.where(ix, t, f)
            output += "".join(row) + "\n"
        return output

    def __str__(self):
        """Make a textual representation of the image."""
        return repr(self._mask)

    def __bool__(self):
        """Check whether any of the mask elements are set."""
        return np.any(self._mask)

    def __invert__(self):
        """Invert the mask."""
        return np.logical_not(self._mask)

    def __neg__(self):
        """Invert the mask."""
        return np.logical_not(self._mask)

    def _repr_png_(self):
        """Provide a display function for iPython/Jupyter."""
        fig = imshow(self._mask.astype(int))
        data = StreamIO()
        fig.savefig(data, format="png")
        plt.close(fig)
        data.seek(0)
        ret = data.read()
        data.close()
        return ret

    def clear(self):
        """Clear a mask."""
        self._imagearray.mask = np.zeros_like(self._imagearray)

    def invert(self):
        """Invert the mask."""
        self._imagearray.mask = ~self._imagearray.mask

    def select(self, **kargs):
        """Interactive selection mode.

        This method allows the user to interactively choose a mask region on the image. It will require the
        Matplotlib backen to be set to Qt or other non-inline backend that supports a user vent loop.

        The image is displayed in the window and athe user can interact with it with the mouse and keyboard.

            -   left-clicking the mouse sets a new vertex
            -   right-clicking the mouse removes the last set vertex
            -   pressing "i" inverts the mask (i.e. controls whether the shape the user is drawing is masked or clear)
            -   pressing "p" sets polygon mode (the default) - each vertex is then the corener of a polygon. The
                polygon
                vertices are defined in order going around the shape.
            -   pressing "r" sets rectangular mode. The first vertex defined is one corner. With only two vertices the
                rectangle is not-rotated and the two vertices define opposite corners. If three vertices are defined
                then the first two form one side and then third vertex controls the extent of the rectangle in the
                direction perpendicular to the side defined.
            -   pressing "c" sets circle/ellipse mode. The first vertex defines one point on the circumference of the
                circle, the next point will define a point on the opposite side of the circumference. If three
                vertices are defined then a circle that passes through all three of them is used. Defining 4
                vertices causes the mode to attempt to find the non-rotated ellipse through the points and further
                vertices allows the ellipse to be rotated.

        This method directly sets the mask and then returns a copy of the parent :py:class:`Stoner.ImageFile`.
        """
        selection = kargs.get("_selection", [])
        if len(selection) == 0:
            selector = ShapeSelect()
            self._imagearray.mask = selector(self._imagearray)
            selection.append(self._imagearray.mask)
        elif len(selection) == 1 and isinstance(selection[0], np.ndarray) and selection[0].dtype.kind == "b":
            self._imagearray.mask = selection[0]
        else:
            raise ValueError("Unknown value for private keyword _selection")
        return self._imagefolder

    def threshold(self, thresh=None):
        """Mask based on a threshold.

        Keyword Arguments:
            thresh (float):
                Threshold to apply to the current image - default is to calculate using threshold_otsu
        """
        if thresh is None:
            thresh = self._imagearray.threshold_otsu()
        self._imagearray.mask = self._imagearray > thresh



# -*- coding: utf-8 -*-
"""A copy of mpltools.special.errorfill with a few hacks to make it place nice.

This software is licensed under the Modified BSD License.

Copyright (c) 2012, Tony S. Yu
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are met:

    * Redistributions of source code must retain the above copyright
      notice, this list of conditions and the following disclaimer.

    * Redistributions in binary form must reproduce the above copyright
      notice, this list of conditions and the following disclaimer in the
      documentation and/or other materials provided with the distribution.

    * Neither the name "mpltools" nor the names of its contributors may be used
      to endorse or promote products derived from this software without
      specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
DISCLAIMED. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE
OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
"""
__all__ = [
    "errorfill",
    "extrema_from_error_input",
    "fill_between",
    "fill_between_x",
    "hsl2rgb",
    "joy_division",
    "auto_fit_fontsize",
]
import warnings
from distutils.version import LooseVersion

import numpy as np
import matplotlib.pyplot as plt
from matplotlib.transforms import Bbox

from Stoner.compat import mpl_version

__all__ = ["errorfill"]

ONE_SIXTH = 1 / 6
ONE_THIRD = 1 / 3
TWO_THIRD = 2 / 3


def errorfill(
    x,
    y,
    yerr=None,
    xerr=None,
    color=None,
    ls=None,
    lw=None,
    alpha=1,
    alpha_fill=0.5,
    label="",
    label_fill="",
    ax=None,
    **kargs
):
    """Plot data with errors marked by a filled region.

    Args:
        x, y (arrays):
            Coordinates of data.
        yerr, xerr ([scalar | N, (N, 1), or (2, N) array]):
            Error for the input data:
                -   If scalar, then filled region spans `y +/- yerr` or `x +/- xerr`.
        color (Matplotlib color):
            Color of line and fill region.
        ls (Matplotlib line style):
            Style of the line
        lw (Matplotlib line width, float):
            Width of the line in points
        alpha (float):
            Opacity used for plotting.
        alpha_fill (float):
            Opacity of filled region. Note: the actual opacity of the fill is
            `alpha * alpha_fill`.
        label (str):
            Label for line.
        label_fill (str):
            Label for filled region.
        ax (Axis instance):
            The plot is drawn on axis `ax`. If `None` the current axis is used
    """
    ax = ax if ax is not None else plt.gca()

    alpha_fill *= alpha

    if color is None:
        if LooseVersion(mpl_version) < LooseVersion("1.5.0"):
            color = next(ax._get_lines.color_cycle)
        else:
            color = next(ax._get_lines.prop_cycler)["color"]
    if ls is None:
        ls = plt.rcParams["lines.linestyle"]
    if lw is None:
        lw = plt.rcParams["lines.linewidth"]
    ax.plot(x, y, color, linestyle=ls, linewidth=lw, alpha=alpha, label=label, **kargs)

    if yerr is not None and xerr is not None:
        msg = "Setting both `yerr` and `xerr` is not supported. Ignore `xerr`."
        warnings.warn(msg)

    kwargs_fill = dict(color=color, alpha=alpha_fill, label=label_fill)
    if yerr is not None:
        ymin, ymax = extrema_from_error_input(y, yerr)
        if x.size > 1:
            fill_between(x, ymax, ymin, ax=ax, **kwargs_fill)
    elif xerr is not None:
        xmin, xmax = extrema_from_error_input(x, xerr)
        fill_between_x(y, xmax, xmin, ax=ax, **kwargs_fill)


def extrema_from_error_input(z, zerr):
    """Work out where to draw limits."""
    if np.isscalar(zerr) or len(zerr) == len(z):
        zmin = z - zerr
        zmax = z + zerr
    elif len(zerr) == 2:
        zmin, zmax = z - zerr[0], z + zerr[1]
    return zmin, zmax


# Wrappers around `fill_between` and `fill_between_x` that create proxy artists
# so that filled regions show up correctly legends.


def fill_between(x, y1, y2=0, ax=None, **kwargs):
    """Draw shading around line."""
    ax = ax if ax is not None else plt.gca()
    ym = (y1 + y2) / 2.0
    yd = (y1 - y2) / 3.0
    # Remove any bad data points
    keep = np.logical_not(np.isnan(ym))
    x = x[keep]
    ym = ym[keep]
    yd = yd[keep]
    alpha = kwargs["alpha"]
    a = np.linspace(0.1, 0.9, 15)
    z = lambda x, s, y: s * np.sqrt(-2 * np.log(np.sqrt(2 * np.pi) * s * y))
    for h in a:
        y = h / (np.sqrt(2 * np.pi) * yd)
        y1 = ym - (z(y, yd, y))
        y2 = ym + (z(y, yd, y))
        kwargs["alpha"] = alpha * h
        if x.size > 1:
            ax.fill_between(x, y1, y2, **kwargs)
    ax.add_patch(plt.Rectangle((0, 0), 0, 0, **kwargs))


def fill_between_x(x, y1, y2=0, ax=None, **kwargs):
    """Draw shading around line."""
    ax = ax if ax is not None else plt.gca()
    ym = (y1 + y2) / 2.0
    yd = (y1 - y2) / 3.0
    alpha = kwargs["alpha"]
    a = np.linspace(0.1, 0.9, 15)
    z = lambda x, s, y: s * np.sqrt(-2 * np.log(np.sqrt(2 * np.pi) * s * y))
    for h in a:
        y = h / (np.sqrt(2 * np.pi) * yd)
        y1 = ym - (z(y, yd, y))
        y2 = ym + (z(y, yd, y))
        kwargs["alpha"] = alpha * h
        ax.fill_betweenx(x, y1, y2, **kwargs)
    ax.add_patch(plt.Rectangle((0, 0), 0, 0, **kwargs))


def hsl2rgb(hue, sat, lum, alpha=False):
    """Convert from hsl colourspace to rgb colour space with numpy arrays for speed.

    Args:
        hue (array):
            Hue value
        sat (array):
            Saturation value
        lum (array):
            Luminence value

    Returns:
        2D array (Mx3) of unsigned 8bit integers
    """
    hue = np.atleast_1d(hue)
    sat = np.atleast_1d(sat)
    lum = np.atleast_1d(lum)

    Hp = hue * 6 % 6
    C = (1 - np.abs(2 * lum - 1)) * sat
    X = C * (1 - np.abs(Hp % 2 - 1))
    output = np.ones((hue.size, 3))
    zero = np.zeros_like(X)
    select = np.column_stack([Hp, Hp, Hp])
    m = np.column_stack([lum - C / 2, lum - C / 2, lum - C / 2])
    output = np.where(select > 0, np.column_stack([C, X, zero]), output)
    output = np.where(select > 1, np.column_stack([X, C, zero]), output)
    output = np.where(select > 2, np.column_stack([zero, C, X]), output)
    output = np.where(select > 3, np.column_stack([zero, X, C]), output)
    output = np.where(select > 4, np.column_stack([X, zero, C]), output)
    output = np.where(select > 5, np.column_stack([C, zero, X]), output)
    output += m
    output = (output * 255).astype("u8")
    if alpha:
        output = np.append(output, [1.0])
    return output


def joy_division(x, y, z, **kargs):
    """Produce a classic black and white water fall plot.

    Parameters:
        x,y,z (1D arrays):
            x y and z coordinates. data should be arranged so that z(x,y=constant)

    Keyword Parameters:
        ax (matplotlib.Axes):
            Axes to use (defaults to current axes)
        y_shift (float):
            Shift in data for successive y values. Defaults to z-span / number of values in y
        bg_colour (matplotlib colour):
            Colour to use for background of the plot (default is "k" for black)
        colour (matplotlib colour):
            Colour of the lines on the plot (default 'white')
        axes_colour (matplotlib colour):
            Coulour of the frame, ticks, labels on the plot (default, same as *colour*)
        linewidth (float):
            Width of lines to use on the plot (default 2)
        legend_fmt (str):
            String to use to format the elgend text. Should include one place holder {}. Default "{}"

    Returns:
        None

    Constructors a mono-chromatic waterfall plot in the style of the Joy Division album cover of Pulsar signals.
    """
    ax = kargs.pop("ax", plt.gca())
    y_shift = kargs.pop("y_shift", (z.max() - z.min()) / np.unique(y).size)
    bg_colour = kargs.pop("bg_color", "k")
    color = kargs.pop("color", kargs.pop("colour", "w"))
    axes_colour = kargs.pop("axes_color", color)
    lw = kargs.pop("linewidth", 2)
    legend_fmt = kargs.pop("legend_fmt", "{}")

    ax.figure.set_facecolor(bg_colour)
    ax.set_facecolor(bg_colour)

    yvals = np.unique(y)
    yvals = np.sort(yvals)

    data = np.column_stack((x, z, y))

    for ix, yval in enumerate(yvals):
        offset = y_shift * (len(yvals) - ix - 1)
        this_data = data[y == yval, :]

        ax.plot(
            this_data[:, 0],
            this_data[:, 1] + offset,
            color,
            lw=lw,
            zorder=(ix + 1) * 2,
            label=legend_fmt.format(yval),
        )
        ax.fill_between(
            this_data[:, 0], this_data[:, 1] + offset, offset, facecolor=bg_colour, lw=0, zorder=(ix + 1) * 2 - 1
        )

    ax.tick_params(color=axes_colour, labelcolor=axes_colour)
    for spine in ax.spines.values():
        spine.set_edgecolor(axes_colour)
    ax.set_title(ax.get_title(), color=axes_colour)
    ax.set_xlabel(ax.get_xlabel(), color=axes_colour)
    ax.set_ylabel(ax.get_ylabel(), color=axes_colour)

    # plt.draw()

    # for t in ax.get_legend().get_texts():
    #     t.set_color(axes_colour)
    # plt.draw()


def auto_fit_fontsize(text, width, height, scale_down=True, scale_up=False):
    """Resale the font size of a matplotlib text object to fit within a box.

    Arguments:
        text (matplotlib.text.Text):
            Text object to be scaled in Figure units.
        width,height (float):
            Target width and height to scale to.

    Keyword Arguments:
        scale_down, scale_up (bool):
            Whether to reduce the font size to fit (default True), or increase it to fit (default False)

    Returns:
        (float):
            scaling factor applied.
    """
    fig = text.axes.figure

    # get text bounding box in figure coordinates
    renderer = fig.canvas.get_renderer()
    bbox_text = text.get_window_extent(renderer=renderer)

    # transform bounding box to data coordinates
    bbox_text = Bbox(fig.transFigure.inverted().transform(bbox_text))
    text_width, text_height = bbox_text.width, bbox_text.height

    scale_w = abs(width / text_width) if text_width != 0 else 1.0
    scale_h = abs(height / text_height) if text_height != 0 else 1.0

    scale = 1.0
    if scale_down:
        scale = min(scale_w, scale_h, scale)
    if scale_up:
        scale = max(scale_w, scale_h, scale)
    if not np.isclose(scale, 1.0):
        text.set_fontsize(text.get_fontsize() * scale)
    return scale



# -*- coding: utf-8 -*-
"""Stoner.Utils - a module of some slightly experimental routines that use the Stoner classes."""
__all__ = ["split_up_down", "ordinal", "hysteresis_correct"]
import numpy as np
from numpy import max, argmax, mean  # pylint: disable=redefined-builtin
from scipy.stats import sem
from scipy.optimize import fsolve

from .Core import DataFile

from .tools import format_error, make_Data, ordinal
from . import DataFolder
from .analysis.fitting.models.generic import linear


def _step(x, m, c, h):
    """Provide a sloping step function for fitting to the extrema of a hysteresis loop.

    Args:
        x (array-like):
            Field (x) data of loop.
        m (float):
            susceptibility(slopee) of loop.
        c (float):
            vertical offset of loop.
        h (float):
            Saturatted moement (height) of loop.

    Returns:
        y (Tarray-like):
            Calculated moment of loop.
    """
    mid = (x.max() + x.min()) / 2.0

    X = x - mid
    y = m * x + c + np.sign(X) * h
    return y


def _h_sat_linear(d, i, Ms_vals, Hsat_vals, h_sat_fraction):
    """Determine the saturation field from linear fits to the hysteresis loop.

    This method uses the intercept of the saturated state with zero field portions of the loop.
    """
    from Stoner.analysis.fitting.models.generic import Linear

    Ms, Ms_err, _ = Ms_vals
    Hsat, Hsat_err = Hsat_vals

    # Fit a straight line to the central fraction of the data
    if i == 1:
        bounds = lambda x, r: np.abs(r.y) < np.abs(Ms) * h_sat_fraction
    else:
        bounds = lambda x, r: np.abs(r.y) < np.abs(Ms) * h_sat_fraction
    popt, pcov = d.lmfit(Linear, bounds=bounds)
    perr = np.sqrt(np.diag(pcov))
    popt = popt[:2]
    pferr = perr / popt
    # Pick the positive or negative Ms depending on the up or down loop
    Ms_i = Ms if i == 0 else -Ms
    # Find the intercept
    Hsat[1 - i] = fsolve(lambda x: Linear().func(x, *popt) - Ms_i, 0)[0]
    # Uncertainty is the sum of error in slope * found intercept and error in Ms times slope
    Hsat_err[1 - i] = np.sqrt((Hsat[1 - i] * pferr[1]) ** 2 + (popt[1] * Ms_err) ** 2)

    return (Hsat, Hsat_err)


def _h_sat_susceptibility(d, i, Ms_vals, Hsat_vals, h_sat_fraction):  # pylint: disable=unused-argument
    """Determine the saturation field from the change in local sysceptibility in the loop."""
    Hsat, Hsat_err = Hsat_vals
    xi = d.SG_Filter(order=1)[0, :]
    m, h = d.SG_Filter()
    tmp = np.column_stack((h, m, xi))[4:-4]
    tmp = make_Data(tmp, setas="x.y")
    threshold = tmp.mean(bounds=lambda r: not 7 < r.i < len(tmp) - 7) * h_sat_fraction
    hs = tmp.threshold(threshold, all_vals=True, rising=i != 0, falling=i == 0)  # Get the H_sat value
    Hsat[1 - i] = mean(hs)  # Get the H_sat value
    if hs.size > 1:
        Hsat_err[1 - i] = sem(hs)
    else:
        Hsat_err[1 - i] = np.NaN
    return (Hsat, Hsat_err)


def _h_sat_delta_M(d, i, Ms_vals, Hsat_vals, h_sat_fraction):  # pylint: disable=unused-argument
    """Determine the saturation field from the change in magnetisation from Ms."""
    m_sat = Ms_vals[2]
    Hsat, Hsat_err = Hsat_vals
    m, h = d.SG_Filter()
    tmp = np.column_stack((h, m))[4:-4]
    tmp = make_Data(tmp, setas="xy")
    threshold = m_sat[i]
    hs = tmp.threshold(threshold, all_vals=True, rising=True, falling=True).view(np.ndarray)  # Get the H_sat value
    if hs.size > 1:
        hs = hs[np.argmin(np.abs(hs))]
    Hsat[1 - i] = hs  # Get the H_sat value
    Hsat_err[1 - i] = np.NaN
    return (Hsat, Hsat_err)


def _up_down(data):
    """Split data d into rising and falling sections and then add and sort the two sets.

    This routine searches for the local maxima and minima in the x-axis data by identifying points
    where the x-data is more than 95% of the total span from the mid-point. It thenuses this to identify
    ranges of rows where the x value passes from one extreme to the other and splits the data file up on
    that basis. This assumes that the x-data is reasonably well behaved in the the increments between
    successive x-values are larger than the noise in x. This will work best if the x value is derived from a set value
    rather than the read-back value.

    Args:
        data (Data):
            DataFile like object with x and y columns set

    Returns:
        (Data, Data):
            Tuple of two DataFile like instances for the rising and falling data.
    """
    # Calculate x span and mid-point for working out limits to search for maxima.
    lx, hx = data.span(data.setas.xcol)
    mid = (lx + hx) / 2.0
    span = hx - lx
    high = data.x > mid + 0.45 * span
    low = data.x < mid - 0.45 * span

    # Locate points where we cross a threshold
    t = np.zeros((2, len(data) + 2), dtype=bool)
    t[0, 1:-1] = high
    t[1, 1:-1] = low
    t = np.diff(t)

    # Find  the indiices of the highest and lowest extrema
    high_i = np.arange(len(data) + 1, dtype=int)[t[0, :]]
    low_i = np.arange(len(data) + 1, dtype=int)[t[1, :]]
    if low_i.size > 2:
        low_i = np.reshape(low_i, (2, -1)).mean(axis=1)
    else:
        low_i = np.array(low_i.mean())
    if high_i.size > 2:
        high_i = np.reshape(high_i, (2, -1)).mean(axis=1)
    else:
        high_i = np.array(high_i.mean())
    # Build a sorted list of extrema positions + the start and end of the data
    indices = np.unique(np.append(low_i, np.append(high_i, np.array([0, len(data) - 1]))))
    indices = np.ceil(indices).astype(int)
    indices.sort()
    if indices[-1] >= len(data):  # Remove possible over-range element
        indices = indices[0:-1]

    # Build a boolean array to index the data for rows that where x is increasing
    rising = np.zeros(len(data), dtype=bool)
    for ix, iy in zip(indices[:-1], indices[1:]):
        if data.x[iy] > data.x[ix]:
            rising[ix:iy] = True

    # Clone the data and select rows that are either rising or falling in the two clones.
    up = data.clone
    up.data = up.data[rising]
    down = data.clone
    down.data = down.data[~rising]
    # Done.
    return up, down


def split_up_down(data, col=None, folder=None):
    """Split the DataFile data into several files where the column *col* is either rising or falling.

    Args:
        data (:py:class:`Stoner.Core.DataFile`):
            object containign the data to be sorted
        col (index):
            is something that :py:meth:`Stoner.Core.DataFile.find_col` can use
        folder (:py:class:`Stoner.Folders.DataFolder` or None):
            if this is an instance of :py:class:`Stoner.Folders.DataFolder` then add
            rising and falling files to groups of this DataFolder, otherwise create a new one

    Returns:
        (:py:class:`Sonter.Folder.DataFolder`):
            with two groups, rising and falling
    """
    a = make_Data(data)
    if col is None:
        _ = a._col_args()
        col = _.xcol
    width = int(len(a) / 10)
    if width % 2 == 0:  # Ensure the window for Satvisky Golay filter is odd
        width += 1
    setas = a.setas.clone
    a.setas = ""
    peaks = list(a.peaks(ycol=col, width=width, full_data=False))
    troughs = list(a.peaks(ycol=col, width=width, peaks=False, troughs=True, full_data=False))
    a.setas = setas
    if peaks and troughs:  # Ok more than up down here
        order = peaks[0] < troughs[0]
    elif peaks:  # Rise then fall
        order = True
    elif troughs:  # Fall then rise
        order = False
    else:  # No peaks or troughs so just return a single rising
        ret = DataFolder(readlist=False)
        ret += data
        return ret
    splits = [0, len(a)]
    splits.extend(peaks)
    splits.extend(troughs)
    splits.sort()
    splits = [int(s) for s in splits]
    if not isinstance(folder, DataFolder):  # Create a new DataFolder object
        output = DataFolder(readlist=False)
    else:
        output = folder
    output.add_group("rising")
    output.add_group("falling")

    if order:
        risefall = ["rising", "falling"]
    else:
        risefall = ["falling", "rising"]
    for i in range(len(splits) - 1):
        working = data.clone
        working.data = data.data[splits[i] : splits[i + 1], :]
        output.groups[risefall[i % 2]].append(working)
    return output


Hickeyify = format_error


def hysteresis_correct(data, **kargs):
    """Perform corrections to a hysteresis loop.

    Args:
        data (Data):
            The data containing the hysteresis loop. The :py:attr:`DataFile.setas` attribute
            should be set to give the H and M axes as x and y.

    Keyword Arguments:
        correct_background (bool):
            Correct for a diamagnetic or paramagnetic background to the hystersis loop
            also recentres the loop about zero moment (default True).
        correct_H (bool):
            Finds the co-ercive fields and sets them to be equal and opposite. If the loop is sysmmetric
            this will remove any offset in filed due to trapped flux (default True)
        saturated_fraction (float):
            The fraction of the horizontal (field) range where the moment can be assumed to be
            fully saturated. If an integer is given it will use that many data points at the end of the loop.
        h_sat_method (str):
            The method used to determine thwe saturation field. Options are -
            -   "linear_intercept" (default): Fit a straight line to the central region of each branch of the loop
                and look at the
                intercept with the relevant saturation moment.
            -   "delta_M": Look for a field where the moment has changed by *h_sat_fraction* times the error in M_s.
            -   "susceptibility" - Calculate H_sat from where the susceptibility changes by 1% of the average
                susceptibility
        h_sat_fraction (float):
            The central fraction of the saturation moment that is used for calculating the saturation field. Defaults
            to 0.5
        xcol (column index):
            Column with the x data in it
        ycol (column_index):
            Column with the y data in it
        setas (string or iterable):
            Column assignments.

    Returns:
        (:py:class:`Stoner.Data`):
            The original loop with the x and y columns replaced with corrected data and extra metadata added to give
            the background suceptibility, offset in moment, co-ercive fields and saturation magnetisation.
    """
    if isinstance(data, DataFile):
        cls = type(data)
    else:
        cls = make_Data(None)
    data = cls(data)

    # Get other keyword arguments
    correct_background = kargs.pop("correct_background", True)
    correct_H = kargs.pop("correct_H", True)
    saturation_fraction = kargs.pop("saturated_fraction", 0.2)

    h_sat_method = kargs.pop("h_sat_method", "linear_intercept")
    h_sat_fraction = kargs.pop("h_sat_fraction", 0.5 if h_sat_method == "linear_intercept" else 2.0)
    hsat_methods = {
        "linear_intercept": _h_sat_linear,
        "susceptibility": _h_sat_susceptibility,
        "delta_M": _h_sat_delta_M,
    }
    if callable(h_sat_method) or h_sat_method in hsat_methods:
        h_sat_method = hsat_methods.get(h_sat_method, h_sat_method)
    else:
        raise ValueError("Saturation field method not recognized!")

    for k in kargs:
        try:
            setattr(data, k, kargs[k])
        except AttributeError:
            if data.debug:
                print("Error setting attribute from keyword {k}={args[k]}")
    if "setas" in kargs:  # Allow us to override the setas variable
        data.setas = kargs.pop("setas")

    xcol = kargs.pop("xcol", data.setas.xcol)
    ycol = kargs.pop("ycol", data.setas.ycol)
    # Get xcol and ycols from kargs if specified
    _ = data._col_args(xcol=xcol, ycol=ycol)
    data.setas(x=_.xcol, y=_.ycol, reset=False)
    # Split into two sets of data:

    up, down = _up_down(data)

    mid = (data.x.max() + data.x.min()) / 2.0
    span = data.x.max() - data.x.min()
    low = mid - span * (1 - saturation_fraction) / 2
    high = mid + span * (1 - saturation_fraction) / 2

    popt, pcov = data.curve_fit(_step, bounds=lambda x, r: not low < x < high)

    perr = np.sqrt(np.diag(pcov))

    Ms = popt[2]
    Ms_err = perr[2]

    data["Ms"] = Ms  # mean(Ms)
    data["Ms Error"] = Ms_err
    data["Offset Moment"] = popt[1]
    data["Offset Moment Error"] = perr[1]
    data["Background susceptibility"] = popt[0]
    data["Background Susceptibility Error"] = perr[0]

    if correct_background:
        fixes = [data, up, down]
    else:
        fixes = [up, down]
    m = popt[0]
    c = popt[1] - m * mid

    for d in fixes:
        d.y = d.y - linear(d.x, c, m)

    Hc = [None, None]
    Hc_err = [None, None]
    Hsat = [None, None]
    Hsat_err = [None, None]
    Mr = [None, None]
    Mr_err = [None, None]

    m_sat = [Ms - h_sat_fraction * Ms_err, -Ms + h_sat_fraction * Ms_err]
    single_side = False

    for i, d in enumerate([up, down]):
        if len(d) < 2:
            single_side = True
            continue
        hc = d.threshold(0.0, all_vals=True, rising=i == 0, falling=i != 0)  # Get the Hc value
        Hc[i] = mean(hc)
        if hc.size > 1:
            Hc_err[i] = sem(hc)

        Hsat, Hsat_err = h_sat_method(d, i, (Ms, Ms_err, m_sat), (Hsat, Hsat_err), h_sat_fraction)

        mr = d.threshold(0.0, col=_.xcol, xcol=_.ycol, all_vals=True, rising=True, falling=True)
        Mr[i] = mean(mr)
        if mr.size > 1:
            Mr_err[i] = sem(mr)

    if correct_H and not single_side:
        Hc_mean = mean(Hc)
        for d in [data, up, down]:
            d.x = d.x - Hc_mean
        data["Exchange Bias offset"] = Hc_mean
    else:
        Hc_mean = 0.0

    if not single_side:
        data["Hc"] = (Hc[1] - Hc_mean, Hc[0] - Hc_mean)
        data["Hc_mean"] = np.abs(np.array(Hc)).mean()
        data["Hsat"] = (Hsat[1] - Hc_mean, Hsat[0] - Hc_mean)
        data["Hsat Error"] = (Hsat_err[1], Hsat_err[0])
        data["Hsat_mean"] = np.abs(np.array(Hsat)).mean()
        data["Hsat_mean Error"] = np.sqrt(np.array(Hsat_err) ** 2)[0] / 2.0
    else:
        data["Hc"] = [x for x in Hc if x is not None]
        data["Hc_mean"] = abs(data["Hc"][0])
        data["Hsat"] = [x for x in Hsat if x is not None]
        data["Hsat Error"] = [x for x in Hsat_err if x is not None]
        data["Hsat_mean"] = abs(data["Hsat"][0])
        data["Hsat_mean Error"] = abs(data["Hsat Error"][0])
    data["Remenance"] = Mr

    bh = (-data.x) * data.y
    i = argmax(bh)
    data["BH_Max"] = max(bh)
    data["BH_Max_H"] = data.x[i]

    data["Area"] = data.integrate(output="result")
    return cls(data)



# -*- coding: utf-8 -*-
"""Utility functions to support :py:class:`Stoner.folders.core.objectFolder` operations."""
__all__ = [
    "pathsplit",
    "pathjoin",
    "scan_dir",
    "discard_earlier",
    "filter_files",
    "get_pool",
    "removeDisallowedFilenameChars",
]
import os.path as path
import re
import string
import fnmatch
import pathlib
from multiprocessing.pool import ThreadPool

from numpy import array
import multiprocess as multiprocessing

from Stoner.compat import string_types, _pattern_type
from Stoner.tools import get_option


def pathsplit(pth):
    """Split pth into a sequence of individual parts with path.split."""
    pth = pathlib.Path(pth)
    ret = [pth.name]
    ret.extend([x.name for x in pth.parents])
    ret.reverse()
    return [str(x) for x in ret if x != ""]


def pathjoin(*args):
    """Join a path like path.join, but then replace the path separator with a standard /."""
    if len(args) > 1:
        tmp = path.join(args[0], *args[1:])
        return tmp.replace(path.sep, "/")


def scan_dir(root):
    """Gather a list of files and directories."""
    dirs = []
    files = []
    root = pathlib.Path(root)
    for f in root.glob("*"):
        if f.is_dir():
            dirs.append(f.name)
        elif f.is_file():
            files.append(f.name)
    return dirs, files


def discard_earlier(files):
    """Discard files where a similar named file with !#### exists."""
    search = re.compile(r"^(?P<basename>.*)\!(?P<rev>\d+)(?P<ext>\.[^\.]*)$")
    dups = dict()
    ret = []
    for f in files:
        match = search.match(f)
        if match:
            fname = f"{match.groupdict()['basename']}{match.groupdict()['ext']}"
            rev = int(match.groupdict()["rev"])
            entry = dups.get(fname, [])
            entry.append((rev, f))
            dups[fname] = entry
        else:
            entry = dups.get(f, [])
            entry.append((-1, f))
            dups[f] = entry
    for f, revs in dups.items():
        rev = sorted(revs)[-1]
        ret.append(rev[1])
    return ret


def filter_files(files, patterns, keep=True):
    """Filter a list of files against include/exclusion patterns.

    Args:
        files (list of str): Filename to filter
        pattens (list of (str,regular expressions): List of patterns to consider

    Keyword Arguments:
        keep (bool): True (default) to keep matching files.

    Returns:
        (list of str): Files that pass the filter
    """
    dels = []
    for p in patterns:  # Remove excluded files
        if isinstance(p, string_types):
            for f in list(fnmatch.filter(files, p)):
                dels.append(files.index(f))
        if isinstance(p, _pattern_type):
            # For reg expts we iterate over all files, but we can't delete matched
            # files as we go as we're iterating over them - so we store the
            # indices and delete them later.
            for f in files:
                if p.search(f):
                    dels.append(files.index(f))
        index = array([not keep ^ (i in dels) for i in range(len(files))], dtype=bool)
        files = (array(files)[index]).tolist()
    return files


def get_pool(_serial=False):
    """Get a Pool and map implementation depending on options.

    Returns:
        Pool(),map: Pool object if possible and map implementation.
    """
    if get_option("multiprocessing") and not _serial:
        try:
            if get_option("threading"):
                p = ThreadPool(processes=int(multiprocessing.cpu_count() - 1))
            else:
                p = multiprocessing.Pool(int(multiprocessing.cpu_count() / 2))
            imap = p.imap
        except (ArithmeticError, AttributeError, LookupError, RuntimeError, NameError, OSError, TypeError, ValueError):
            # Fallback to non-multiprocessing if necessary
            p = None
            imap = map
    else:
        p = None
        imap = map
    return p, imap


def removeDisallowedFilenameChars(filename):
    """Clean characters in filenames.

    Args:
        filename (string): filename to cleanse

    Returns:
        A filename with non ASCII characters stripped out
    """
    validFilenameChars = "-_.() %s%s" % (string.ascii_letters, string.digits)
    return "".join([c for c in filename if c in validFilenameChars])



"""Sub package of various built-in models for the Stoner package."""

__all__ = [
    "generic",
    "thermal",
    "magnetism",
    "tunnelling",
    "e_transport",
    "superconductivity",
    "_get_model_",
    "make_model",
    "cfg_data_from_ini",
    "cfg_model_from_ini",
]

from collections.abc import Mapping
from configparser import ConfigParser as SafeConfigParser
from functools import wraps
from io import IOBase

import numpy as np
from lmfit import Model
from lmfit.models import update_param_vals

from Stoner.compat import string_types
from Stoner.tools import make_Data
import Stoner.Core as _SC_
from . import generic, thermal, magnetism, tunnelling, e_transport, superconductivity


def _get_model_(model):
    """Utility meothd to manage creating an lmfit.Model.

    Args:
        model (str, callable, Model):
            The model to be setup.

    Returns:
        An llmfit.Model instance

    model can be of several different types that determine what to do:
        -   A string. In which ase it should be a fully qualified name of a function or class to be imported.
            The part after the final period will be assumed to be the name and the remainder the module to be
            imported.
        -   A callable object. In this case the callable will be passed to the constructor of Model and a fresh
            Model instance is constructed
        -   A subclass of lmfit.Model - in which case it is instantiated.
        -   A Model instance - in which case no further action is necessary.
    """
    if isinstance(model, string_types):  # model is a string, so we;ll try importing it now
        parts = model.split(".")
        model = parts[-1]
        module = ".".join(parts[:-1])
        model = __import__(module, globals(), locals(), (model)).__getattribute__(model)
    if type(model).__name__ == "type" and issubclass(
        model, Model
    ):  # ok perhaps we've got a model class rather than an instance
        model = model()
    if not isinstance(model, Model) and callable(model):  # Ok, wrap the callable in a model
        model = Model(model)
    if not isinstance(model, Model):
        raise TypeError(f"model {model.__name__} is not an instance of llmfit.Model")
    return model


def make_model(model_func):
    """A decorator that turns a function into an lmfit model.

    Args:
        model_func (functrion):
            The function to be wrapped in an lmfit.Model object

    Notes:
        The function being wrapped into the model should have the form::

            def model_func(x_data,*parameters):
                ....

        (i.e. similar to what :py:func:`scipy.optimize.curve_fit` expects). The resulting
        class is a sub-class of :py:class:`lmfit.Model` but also adds a class method
        :py:method:`_ModelDectorator.guesser` which can be used as a decorator to convert another function into
        a :py:meth:`lmfit.Model.guess` method. If using this decorator, the function that does the guessing should
        take the form::

            def guesser_function(y_data,x=x_data,**kargs):
                return (param_1,param_2,....,pram_n)

        Similarly, the class provides a :py:meth:`_ModelDecorator.hinter` decorator which can be used to mark a
        function as something that can generate parameter hints for the model. In this case the function should
        take the form::

            def hinter(**kwargs):
                return {"param_1":{"max":max_val,"min":min_value,"value":start_value},"param_2":{.....}}

        Finally the new model_func class can be instantiated or just passed to :py:meth:`Data.lmfit` etc. directly.
    """

    class _ModelDecorator(Model):
        __doc__ = model_func.__doc__

        def __init__(self, *args, **kargs):
            super().__init__(model_func, *args, **kargs)
            if hasattr(self, "_limits"):
                for param, limit in self._limits().items():
                    self.set_param_hint(param, **limit)
            self.__name__ = self.func.__name__

        def guess(self, y, x=None):  # pylint: disable=unused-argument
            """A default parameter guess method.

            Just guesses 1.0 for everything like :py:func:`scipy.optimize.curve_fit` does."""
            return np.ones(len(self.param_names))

        @classmethod
        def hinter(cls, func):
            """Use the given function to determine the parameter hints.

            Args:
                func (callable):
                    A fimction that rturns a dictionary of dictionaries

            Returns:
                The wrapped hinter function.

            Notes:
                This decorator will modify the instance attributes so that the instance has a method to generate
                parameter hints.

                func should only take keyword arguments as by default it will be called with no arguments during
                model initialisation.
            """

            @wraps(func)
            def _limits_proxy(self, **kargs):
                limits = func(**kargs)
                for param in limits:
                    if param not in self.param_names:
                        raise RuntimeError(f"Unrecognised parameter in hinter function: {param}")
                    if not isinstance(limits[param], Mapping):
                        raise RuntimeError(f"Parameter hint for {param} was not a mapping")
                return limits

            cls._limits = _limits_proxy
            return _limits_proxy

        @classmethod
        def guesser(cls, func):
            """Use the given function as the guess method.

            Args:
                func (callable):
                    A function that guesses the parameter values

            Returns:
                The wrapped guess function.

            Notes:
                This decorator will modify the instance attributes so that the instance has a working guess method.

                func should take at least one positional argument, being the y-data values used to guess parameters.
                It should return a list, tuple of guesses parameter values with one entry for each parameter in
                the model.
            """

            @wraps(func)
            def guess_proxy(self, *args, **kargs):
                """A magic proxy call around a function to guess initial parameters."""
                guesses = func(*args, **kargs)
                pars = {x: y for x, y in zip(self.param_names, guesses)}
                pars = self.make_params(**pars)
                return update_param_vals(pars, self.prefix, **kargs)

            cls.guess = guess_proxy
            return guess_proxy

    return _ModelDecorator


def cfg_data_from_ini(inifile, filename=None, **kargs):
    """Read an inifile and load and configure a DataFile from it.

    Args:
        inifile (str or file):
            Path to the ini file to be read.

    Keyword Arguments:
        filename (string,boolean or None):
            File to load that contains the data.
        **kargs:
            All other keywords are passed to the Data constructor

    Returns:
        An instance of :py:class:`Stoner.Core.Data` with data loaded and columns configured.

    The inifile should contain a [Data] section that contains the following keys:
        -  **type (str):** optional name of DataFile subclass to import.
        -  **filename (str or boolean):** optionally used if *filename* parameter is None.
        - **xcol (column index):** defines the x-column data for fitting.
        - **ycol (column index):** defines the y-column data for fitting.
        - **yerr (column index):** Optional column with uncertainty values for the data
    """
    if SafeConfigParser is None:
        raise RuntimeError("Need to have ConfigParser module installed for this to work.")
    config = SafeConfigParser()
    if isinstance(inifile, string_types):
        config.read(inifile)
    elif isinstance(inifile, IOBase):
        if inifile.seekable():
            inifile.seek(0)
        config.read_file(inifile)
    if not config.has_section("Data"):
        raise RuntimeError("Configuration file lacks a [Data] section to describe data.")

    if config.has_option("Data", "type"):
        typ = config.get("Data", "type").split(".")
        typ_mod = ".".join(typ[:-1])
        typ = typ[-1]
        typ = __import__(typ_mod, fromlist=[typ]).__getattribute__(typ)
    else:
        typ = None
    data = make_Data(**kargs)
    if filename is None:
        if not config.has_option("Data", "filename"):
            filename = False
        else:
            filename = config.get("Data", "filename")
            if filename in ["False", "True"]:
                filename = bool(filename)
    data = data.__class__.load(filename, auto_load=False, filetype=typ)
    cols = {"x": 0, "y": 1, "e": None}  # Defaults

    for c in ["x", "y", "e"]:
        if not config.has_option("Data", c):
            pass
        else:
            try:
                cols[c] = config.get("Data", c)
                cols[c] = int(cols[c])
            except ValueError:
                pass
        if cols[c] is None:
            del cols[c]

    data.setas(**cols)  # pylint: disable=not-callable
    return data


def cfg_model_from_ini(inifile, model=None, data=None):
    r"""Utility function to configure an lmfit Model from an inifile.

    Args:
        inifile (str or file):
            Path to the ini file to be read.

    Keyword Arguments:
        model (str, callable, lmfit.Model instance or sub-class or None):
            What to use as a model function.
        data (DataFile):
            If supplied, the details of the parameter hints and labels and units are included in the data's metadata.

    Returns:
        An llmfit.Model,, a 2D array of starting values for each parameter

    model can be of several different types that determine what to do:
        -   A string. In which ase it should be a fully qualified name of a function or class to be imported.
            The part after the final period will be assumed to be the name and the remainder the module to be
            imported.
        -   A callable object. In this case the callable will be passed to the constructor of Model and a fresh
            Model instance is constructed
        -   A subclass of lmfit.Model - in which case it is instantiated.
        -   A Model instance - in which case no further action is necessary.

    The returned model is configured with parameter hints for fitting with. The second return value is
    a 2D array which lists the starting values for one or more fits. If the inifile describes mapping out
    the :math:`\Chi^2` as a function of the parameters, then this array has a separate row for each iteration.
    """
    config = SafeConfigParser()
    if isinstance(inifile, string_types):
        config.read(inifile)
    elif isinstance(inifile, IOBase):
        if inifile.seekable():
            inifile.seek(0)
        config.read_file(inifile)

    if model is None:  # Check to see if config file specified a model
        try:
            model = config.get("Options", "model")
        except (IOError, ValueError, TypeError) as err:
            raise RuntimeError("Model is notspecifed either as keyword argument or in inifile") from err
    model = _get_model_(model)
    if config.has_option("option", "prefix"):
        prefix = config.get("option", "prefix")
    else:
        prefix = type(model).__name__
    prefix += ":"
    vals = []
    for p in model.param_names:
        if not config.has_section(p):
            raise RuntimeError(f"Config file does not have a section for parameter {p}")
        keys = {
            "vary": bool,
            "value": float,
            "min": float,
            "max": float,
            "expr": str,
            "step": float,
            "label": str,
            "units": str,
        }
        kargs = dict()
        for k in keys:
            if config.has_option(p, k):
                if keys[k] == bool:
                    kargs[k] = config.getboolean(p, k)
                elif keys[k] == float:
                    kargs[k] = config.getfloat(p, k)
                elif keys[k] == str:
                    kargs[k] = config.get(p, k)
        if isinstance(data, _SC_.DataFile):  # stuff the parameter hint data into metadata
            for k in keys:  # remove keywords not needed
                if k in kargs:
                    data[f"{prefix}{p} {k}"] = kargs[k]
            if "lmfit.prerfix" in data:
                data["lmfit.prefix"].append(prefix)
            else:
                data["lmfit.prefix"] = [prefix]
        if "step" in kargs:  # We use step for creating a chi^2 mapping, but not for a parameter hint
            step = kargs.pop("step")
            if "vary" in kargs and "min" in kargs and "max" in kargs and not kargs["vary"]:  # Make chi^2?
                vals.append(np.arange(kargs["min"], kargs["max"] + step / 10, step))
            else:  # Nope, just make a single value step here
                vals.append(np.array(kargs["value"]))
        else:  # Nope, just make a single value step here
            vals.append(np.array(kargs["value"]))
        kargs = {k: kargs[k] for k in kargs if k in ["value", "max", "min", "vary"]}
        model.set_param_hint(p, **kargs)  # set the model parameter hint
    msh = np.meshgrid(*vals)  # make a mesh of all possible parameter values to test
    msh = [m.ravel() for m in msh]  # tidy it up and combine into one 2D array
    msh = np.column_stack(msh)
    return model, msh



"""Provides utility classes for writing simple data analysis scripts more easily.

It has been developed by members of the `Condensed Matter Group<http://www.stoner.leeds.ac.uk/>` at the
`University  of Leeds<http://www.leeds.ac.uk>`.
"""

# pylint: disable=import-error
__all__ = [
    "core",
    "analysis",
    "formats",
    "plot",
    "tools",
    "Image",
    "Data",
    "DataFolder",
    "ImageFile",
    "ImageFolder",
    "ImageStack",
    "set_option",
    "get_option",
    "Options",
]

# These fake the old namespace if you do an import Stoner
from os import path as _path_
import pathlib

from . import core, analysis, formats, plot, tools, Image
from .core.data import Data
from .Folders import DataFolder
from .Image import ImageFile, ImageFolder, ImageStack

from .tools import set_option, get_option, Options as _Options

Options = _Options()


__version_info__ = ("0", "10", "10")
__version__ = ".".join(__version_info__)

__homepath__ = pathlib.Path(__file__).parent.resolve()
__datapath__ = (__homepath__ / ".." / "sample-data").resolve()
__home__ = str(__homepath__)



# -*- coding: utf-8 -*-
"""
test_Core.py
Created on Tue Jan 07 22:05:55 2014

@author: phygbu
"""

import sys
import pathlib
import io
import urllib

from Stoner import Data, __homepath__, __datapath__, ImageFile
from Stoner.Core import DataFile
from Stoner.compat import Hyperspy_ok

import pytest

from Stoner.formats.attocube import AttocubeScan
from Stoner.formats.maximus import MaximusStack
from Stoner.tools.classes import subclasses
from Stoner.core.exceptions import StonerUnrecognisedFormat
from Stoner.formats.facilities import FabioImageFile

pth = __homepath__ / ".."
datadir = __datapath__


def setup_module():
    sys.path.insert(0, str(pth))


def teardown_module():
    sys.path.remove(str(pth))


def list_files():
    skip_files = set([])  # HDF5 loader not working Python 3.5
    incfiles = list(set(datadir.glob("*")) - skip_files)
    incfiles = [x for x in incfiles if x.suffix != ".tdms_index"]
    incfiles = [x for x in incfiles if not x.is_dir()]

    if not Hyperspy_ok:
        print("hyperspy too old, skupping emd file for test")
        incfiles = [x for x in incfiles if not x.name.strip().lower().endswith(".emd")]

    return sorted(incfiles)


listed_files = list_files()


@pytest.mark.parametrize("filename", listed_files)
def test_one_file(tmpdir, filename):
    loaded = Data(filename, debug=False)
    assert isinstance(loaded, DataFile), f"Failed to load {filename.name} correctly."
    if "save" in subclasses()[loaded["Loaded as"]].__dict__:
        pth = pathlib.Path(tmpdir) / filename.name
        _, name, ext = pth.parent, pth.stem, pth.suffix
        pth2 = pathlib.Path(tmpdir) / f"{name}-2{ext}"
        loaded.save(pth, as_loaded=True)
        assert pth.exists() or pathlib.Path(loaded.filename).exists(), f"Failed to save as {pth}"
        pathlib.Path(loaded.filename).unlink()
        loaded.save(pth2, as_loaded=loaded["Loaded as"])
        assert pth2.exists() or pathlib.Path(loaded.filename).exists(), "Failed to save as {}".format(pth)
        pathlib.Path(loaded.filename).unlink()


def test_csvfile():
    csv = Data(
        datadir / "working" / "CSVFile_test.dat", filetype="JustNumbers", column_headers=["Q", "I", "dI"], setas="xye"
    )
    assert csv.shape == (167, 3), "Failed to load CSVFile from text"


def test_attocube_scan(tmpdir):
    tmpdir = pathlib.Path(tmpdir)
    scandir = datadir / "attocube_scan"
    scan1 = AttocubeScan("SC_085", scandir, regrid=False)
    scan2 = AttocubeScan(85, scandir, regrid=False)
    assert scan1 == scan2, "Loading scans by number and string not equal"

    # self.assertEqual(scan1,scan2,"Loading Attocube Scans by root name and number didn't match")

    pth = tmpdir / f"SC_{scan1.scan_no:03d}.hdf5"
    scan1.to_hdf5(pth)

    scan3 = AttocubeScan.read_hdf5(pth)

    assert pth.exists(), f"Failed to save scan as {pth}"
    if scan1 != scan3:
        print("A" * 80)
        print(scan1.layout, scan3.layout)
        for grp in scan1.groups:
            print(scan1[grp].metadata.all_by_keys ^ scan3[grp].metadata.all_by_keys)
    print(scan1.shape)
    assert scan1.layout == scan3.layout, "Roundtripping scan through hdf5 failed"
    pth.unlink()

    pth = tmpdir / f"SC_{scan1.scan_no:03d}.tiff"
    scan1.to_tiff(pth)
    scan3 = AttocubeScan.from_tiff(pth)
    assert pth.exists(), f"Failed to save scan as {pth}"
    if scan1 != scan3:
        print("B" * 80)
        print(scan1.layout, scan3.layout)
        for grp in scan1.groups:
            print(scan1[grp].metadata.all_by_keys ^ scan3[grp].metadata.all_by_keys)
    assert scan1.layout == scan3.layout, "Roundtripping scan through tiff failed"
    pth.unlink()

    scan3 = AttocubeScan()
    scan3._marshall(layout=scan1.layout, data=scan1._marshall())
    assert scan1 == scan3, "Recreating scan through _marshall failed."

    scan1["fwd"].level_image(method="parabola", signal="Amp")
    scan1["bwd"].regrid()


def test_maximus_image():
    pths = list((datadir / "maximus_scan").glob("MPI_210127019*.*"))
    assert len(pths) == 2
    for pth in pths:
        img = ImageFile.load(pth)
        assert img.shape == (1000, 1000)
        assert len(img.metadata) == 196


def test_maximus_stack(tmpdir):
    tmpdir = pathlib.Path(tmpdir)
    scandir = datadir / "maximus_scan" / "MPI_210127021"
    stack = MaximusStack(scandir / "MPI_210127021")
    stack.to_hdf5(tmpdir / "MPI_210127021.hdf5")
    stack2 = MaximusStack.read_hdf5(tmpdir / "MPI_210127021.hdf5")
    assert stack2.shape == stack.shape, "Round trip through MaximusStack"


def test_fail_to_load():
    with pytest.raises(StonerUnrecognisedFormat):
        _ = Data(datadir / "TDMS_File.tdms_index")


def test_arb_class_load():
    _ = Data(datadir / "TDI_Format_RT.txt", filetype="dummy.ArbClass")


def test_url_load():
    """Test URL scheme openers."""
    t1 = Data("https://github.com/stonerlab/Stoner-PythonCode/raw/master/sample-data/hairboRaman.spc")
    assert t1 == Data(__datapath__ / "hairboRaman.spc")
    t2 = Data("https://github.com/stonerlab/Stoner-PythonCode/raw/master/sample-data/New-XRay-Data.dql")
    assert t2 == Data(__datapath__ / "New-XRay-Data.dql")
    resp = urllib.request.urlopen(
        "https://github.com/stonerlab/Stoner-PythonCode/raw/master/sample-data/New-XRay-Data.dql"
    )
    t3 = Data(resp)
    assert t3 == t2


def test_from_bytes():
    """Test loading a binary file as bytes."""
    with open(__datapath__ / "harribo.spc", "rb") as data:
        d = Data(data.read())
    assert d == Data(__datapath__ / "harribo.spc")


def test_from_StringIO():
    """Test loading a binary file as bytes."""
    with open(__datapath__ / "RASOR.dat", "r") as data:
        buffer = io.StringIO(data.read())
    assert Data(buffer) == Data(__datapath__ / "RASOR.dat")


def test_ImageAutoLoad():
    """Test ImageFile autoloading"""
    img = ImageFile(__datapath__ / "kermit.png")
    assert img.shape == (479, 359)
    img = ImageFile(__datapath__ / "working" / "hydra_0017.edf")
    assert img.shape == (512, 768)
    img = ImageFile(__datapath__ / "working" / "Sample_Image_2017-06-03_035.hdf5")
    assert img.shape == (80, 300)


def test_FabioImageFle():
    loader = FabioImageFile()
    loader._load(datadir / "working" / "hydra_0017.edf")
    assert loader.shape == (512, 768)


if __name__ == "__main__":  # Run some tests manually to allow debugging
    pytest.main(["--pdb", __file__])



# -*- coding: utf-8 -*-
"""HDF5 format Data loader routines"""
import os

import h5py
import numpy as np
from importlib import import_module

from ...compat import string_types, bytes2str, path_types, str2bytes
from ...core.exceptions import StonerLoadError
from ..decorators import register_loader, register_saver
from ...tools.file import HDFFileManager


@register_loader(
    patterns=[(".hdf", 16), (".hf5", 16)],
    mime_types=[("application/x-hdf", 16), ("application/x-hdf5", 16)],
    name="HDF5File",
    what="Data",
)
def load_hdf(new_data, filename, *args, **kargs):  # pylint: disable=unused-argument
    """Create a new HDF5File from an actual HDF file."""
    with HDFFileManager(filename, "r") as f:
        data = f["data"]
        if np.prod(np.array(data.shape)) > 0:
            new_data.data = data[...]
        else:
            new_data.data = [[]]
        metadata = f.require_group("metadata")
        typehints = f.get("typehints", None)
        if not isinstance(typehints, h5py.Group):
            typehints = {}
        else:
            typehints = typehints.attrs
        if "column_headers" in f.attrs:
            new_data.column_headers = [bytes2str(x) for x in f.attrs["column_headers"]]
            if isinstance(new_data.column_headers, string_types):
                new_data.column_headers = new_data.metadata.string_to_type(new_data.column_headers)
            new_data.column_headers = [bytes2str(x) for x in new_data.column_headers]
        else:
            raise StonerLoadError("Couldn't work out where my column headers were !")
        for i in sorted(metadata.attrs):
            v = metadata.attrs[i]
            t = typehints.get(i, "Detect")
            if isinstance(v, string_types) and t != "Detect":  # We have typehints and this looks like it got exported
                new_data.metadata[f"{i}{{{t}}}".strip()] = f"{v}".strip()
            else:
                new_data[i] = metadata.attrs[i]
        if isinstance(f, h5py.Group):
            if f.name != "/":
                new_data.filename = os.path.join(f.file.filename, f.name)
            else:
                new_data.filename = os.path.realpath(f.file.filename)
        else:
            new_data.filename = os.path.realpath(f.filename)
    return new_data


@register_saver(patterns=[(".hdf", 16), (".hf5", 16)], name="HDF5File", what="Data")
def save_hdf(save_data, filename=None, **kargs):  # pylint: disable=unused-argument
    """Write the current object into  an hdf5 file or group within a file.

    Writes the data in afashion that is compatible with being loaded in again.

    Args:
        filename (string or h5py.Group):
            Either a string, of h5py.File or h5py.Group object into which to save the file. If this is a string,
            the corresponding file is opened for writing, written to and save again.

    Returns
        A copy of the object
    """
    if filename is None:
        filename = save_data.filename
    if filename is None or (isinstance(filename, bool) and not filename):  # now go and ask for one
        filename = save_data.__file_dialog("w")
        save_data.filename = filename
    if isinstance(filename, path_types):
        mode = "r+" if os.path.exists(filename) else "w"
    else:
        mode = ""
    compression = "gzip"
    compression_opts = 6
    with HDFFileManager(filename, mode) as f:
        f.require_dataset(
            "data",
            data=save_data.data,
            shape=save_data.data.shape,
            dtype=save_data.data.dtype,
        )
        metadata = f.require_group("metadata")
        typehints = f.require_group("typehints")
        for k in save_data.metadata:
            try:
                typehints.attrs[k] = save_data.metadata._typehints[k]
                metadata.attrs[k] = save_data[k]
            except TypeError:
                # We get this for trying to store a bad data type - fallback to metadata export to string
                parts = save_data.metadata.export(k).split("=")
                metadata.attrs[k] = "=".join(parts[1:])
        f.attrs["column_headers"] = [str2bytes(x) for x in save_data.column_headers]
        f.attrs["filename"] = save_data.filename
        f.attrs["type"] = type(save_data).__name__
        f.attrs["module"] = type(save_data).__module__
    return save_data


def _scan_SLS_meta(new_data, group):
    """Scan the HDF5 Group for attributes and datasets and sub groups and recursively add them to the metadata."""
    root = ".".join(group.name.split("/")[2:])
    for name, thing in group.items():
        parts = thing.name.split("/")
        name = ".".join(parts[2:])
        if isinstance(thing, h5py.Group):
            _scan_SLS_meta(new_data, thing)
        elif isinstance(thing, h5py.Dataset):
            if thing.ndim > 1:
                continue
            if np.prod(thing.shape) == 1:
                new_data.metadata[name] = thing[0]
            else:
                new_data.metadata[name] = thing[...]
    for attr in group.attrs:
        new_data.metadata[f"{root}.{attr}"] = group.attrs[attr]


@register_loader(
    patterns=[(".hdf", 16), (".hdf5", 16)],
    mime_types=[("application/x-hdf", 16), ("application/x-hdf5", 16)],
    name="SLS_STXMFile",
    what="Data",
)
def load_sls_stxm(new_data, filename, *args, **kargs):
    """Load data from the hdf5 file produced by Pollux.

    Args:
        h5file (string or h5py.Group):
            Either a string or an h5py Group object to load data from

    Returns:
        itnew_data after having loaded the data
    """
    new_data.filename = filename
    if isinstance(filename, path_types):  # We got a string, so we'll treat it like a file...
        try:
            f = h5py.File(filename, "r")
        except IOError as err:
            raise StonerLoadError(f"Failed to open {filename} as a n hdf5 file") from err
    elif isinstance(filename, h5py.File) or isinstance(filename, h5py.Group):
        f = filename
    else:
        raise StonerLoadError(f"Couldn't interpret {filename} as a valid HDF5 file or group or filename")
    items = [x for x in f.items()]
    if len(items) == 1 and items[0][0] == "entry1":
        group1 = [x for x in f["entry1"]]
        if "definition" in group1 and bytes2str(f["entry1"]["definition"][0]) == "NXstxm":  # Good HDF5
            pass
        else:
            raise StonerLoadError("HDF5 file lacks single top level group called entry1")
    else:
        raise StonerLoadError("HDF5 file lacks single top level group called entry1")
    root = f["entry1"]
    data = root["counter0"]["data"]
    if np.prod(np.array(data.shape)) > 0:
        new_data.data = data[...]
    else:
        new_data.data = [[]]
    _scan_SLS_meta(new_data, root)
    if "file_name" in f.attrs:
        new_data["original filename"] = f.attrs["file_name"]
    elif isinstance(f, h5py.Group):
        new_data["original filename"] = f.name
    else:
        new_data["original filename"] = f.file.filename

    if isinstance(filename, path_types):
        f.file.close()
    new_data["Loaded from"] = new_data.filename

    if "instrument.sample_x.data" in new_data.metadata:
        new_data.metadata["actual_x"] = new_data.metadata["instrument.sample_x.data"].reshape(new_data.shape)
    if "instrument.sample_y.data" in new_data.metadata:
        new_data.metadata["actual_y"] = new_data.metadata["instrument.sample_y.data"].reshape(new_data.shape)
    new_data.metadata["sample_x"], new_data.metadata["sample_y"] = np.meshgrid(
        new_data.metadata["counter0.sample_x"], new_data.metadata["counter0.sample_y"]
    )
    if "control.data" in new_data.metadata:
        mod = import_module("Stoner.Image")
        ImageArray = getattr(mod, "ImageArray")

        new_data.metadata["beam current"] = ImageArray(new_data.metadata["control.data"].reshape(new_data.data.shape))
        new_data.metadata["beam current"].metadata = new_data.metadata
    new_data.data = new_data.data[::-1]
    return new_data


def _hgx_scan_group(new_data, grp, pth):
    """Recursively list HDF5 Groups."""
    if pth in new_data.seen:
        return None
    new_data.seen.append(pth)
    if not isinstance(grp, h5py.Group):
        return None
    if new_data.debug:
        if new_data.debug:
            print(f"Scanning in {pth}")
    for x in grp:
        if pth == "":
            new_pth = x
        else:
            new_pth = pth + "." + x
        if pth == "" and x == "data":  # Special case for main data
            continue
        if isinstance(grp[x], type(grp)):
            _hgx_scan_group(new_data, grp[x], new_pth)
        elif isinstance(grp[x], h5py.Dataset):
            y = grp[x][...]
            new_data[new_pth] = y
    return None


def _hgx_main_data(new_data, data_grp):
    """Work through the main data group and build something that looks like a numpy 2D array."""
    if not isinstance(data_grp, h5py.Group) or data_grp.name != "/current/data":
        raise StonerLoadError("HDF5 file not in expected format")
    root = data_grp["datasets"]
    for ix in root:  # Hack - iterate over all items in root, but actually data is in Groups not DataSets
        dataset = root[ix]
        if isinstance(dataset, h5py.Dataset):
            continue
        x = dataset["x"][...]
        y = dataset["y"][...]
        e = dataset["error"][...]
        new_data &= x
        new_data &= y
        new_data &= e
        new_data.column_headers[-3] = bytes2str(dataset["x_command"][()])
        new_data.column_headers[-2] = bytes2str(dataset["y_command"][()])
        new_data.column_headers[-1] = bytes2str(dataset["error_command"][()])
        new_data.column_headers = [str(ix) for ix in new_data.column_headers]


@register_loader(
    patterns=(".hgx", 16),
    mime_types=[("application/x-hdf", 32), ("application/x-hdf5", 32)],
    name="HGXFile",
    what="Data",
)
def _load(new_data, filename, *args, **kargs):
    """Load a GenX HDF file.

    Args:
        filename (string or bool):
            File to load. If None then the existing filename is used, if False, then a file dialog will be used.

    Returns:
        A copy of the itnew_data after loading the data.
    """
    new_data.filename = filename
    new_data.seen = []
    with HDFFileManager(new_data.filename, "r") as f:
        try:
            if "current" in f and "config" in f["current"]:
                pass
            else:
                raise StonerLoadError("Looks like an unexpected HDF layout!.")
            _hgx_scan_group(new_data, f["current"], "")
            _hgx_main_data(new_data, f["current"]["data"])
        except IOError as err:
            raise StonerLoadError("Looks like an unexpected HDF layout!.") from err
    return new_data



"""Scale data to stitch it together.

This example demonstrates specifying the overlap as an integer and scaling both curves to match the other."""

# pylint: disable=invalid-name
import matplotlib.pyplot as plt

from Stoner import Data
from Stoner.Util import format_error

# Load and plot two sets of data
s1 = Data("Stitch-scan1.txt", setas="xy")
s2 = Data("Stitch-scan2.txt", setas="xy")
s3 = s2.clone
s1.plot(fmt="b-", label="Set 1", linewidth=1)
s2.fig = s1.fig
s2.plot(fmt="r-", label="Set 2", linewidth=1)

# Stitch scan 2 onto scan 1
s2.stitch(s1, overlap=100)

s2.plot(fmt="b:", label="Set 2 Stictched on Set 1", linewidth=1)
s2.title = "Stictching Example"

# Tidy up the plot by adding annotation fo the stirching co-efficients
labels = ["A", "B", "C"]

txt = []
lead = r"$x'\rightarrow x+A$" + "\n" + r"$y'=\rightarrow By+C$" + "\n"
for l, v, e in zip(
    labels, s2["Stitching Coefficients"], s2["Stitching Coefficient Errors"]
):
    txt.append(format_error(v, e, latex=True, prefix=l + "="))
plt.text(1.5, 1.25, lead + "\n".join(txt), fontdict={"size": "x-small"})

# Now stich s1 onto the clone of s2 and plot and label
s1.stitch(s3, overlap=100)
s1.plot(fmt="r:", label="Set 1 sticked on Set 2", linewidth=1)

txt = []
lead = r"$x'\rightarrow x+A$" + "\n" + r"$y'=\rightarrow By+C$" + "\n"
for l, v, e in zip(
    labels, s1["Stitching Coefficients"], s1["Stitching Coefficient Errors"]
):
    txt.append(format_error(v, e, latex=True, prefix=l + "="))
plt.text(5.0, 1.25, lead + "\n".join(txt), fontdict={"size": "x-small"})



"""3D surface plot example."""

# pylint: disable=invalid-name
import numpy as np
import matplotlib.cm

from Stoner import Data

x, y = np.meshgrid(np.linspace(-2, 2, 100), np.linspace(-2, 2, 100))
x = x.ravel()
y = y.ravel()
z = np.cos(4 * np.pi * np.sqrt(x**2 + y**2)) * np.exp(-np.sqrt(x**2 + y**2))

p = Data()
p = p & x & y & z
p.column_headers = ["X", "Y", "Z"]
p.setas = "xyz"

p.plot_xyz(cmap=matplotlib.cm.jet)
p.title = "Surface plot"



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""Implements classes to load file formats from various simulation packages."""

import re

import numpy as np

from ...core.exceptions import StonerLoadError
from ...tools.file import FileManager

from ..decorators import register_loader


def _read_line(data, metadata):
    """Read a single line and add to a metadata dictionary."""
    line = data.readline().decode("utf-8", errors="ignore").strip("#\n \t\r")
    if line == "" or ":" not in line:
        return True
    parts = line.split(":")
    field = parts[0].strip()
    value = ":".join(parts[1:]).strip()
    if field == "Begin" and value.startswith("Data "):
        value = value.split(" ")
        metadata["representation"] = value[1]
        if value[1] == "Binary":
            metadata["representation size"] = value[2]
        return False
    if field not in ["Begin", "End"]:
        metadata[field] = value
    return True


@register_loader(patterns=(".dat", 16), mime_types=("text/plain", 16), name="GenXFile", what="Data")
def load_genx(new_data, filename=None, *args, **kargs):
    """Load function. File format has space delimited columns from row 3 onwards."""
    new_data.filename = filename
    pattern = re.compile(r'# Dataset "([^\"]*)" exported from GenX on (.*)$')
    pattern2 = re.compile(r"#\sFile\sexported\sfrom\sGenX\'s\sReflectivity\splugin")
    i = 0
    ix = 0
    with FileManager(new_data.filename, "r", errors="ignore", encoding="utf-8") as datafile:
        line = datafile.readline()
        match = pattern.match(line)
        match2 = pattern2.match(line)
        if match is not None:
            dataset = match.groups()[0]
            date = match.groups()[1]
            new_data["date"] = date
            i = 2
        elif match2 is not None:
            line = datafile.readline()
            new_data["date"] = line.split(":")[1].strip()
            dataset = datafile.readline()[1:].strip()
            i = 3
        else:
            raise StonerLoadError("Not a GenXFile")
        for ix, line in enumerate(datafile):
            line = line.strip()
            if line in ["# Headers:", "# Column labels:"]:
                line = next(datafile)[1:].strip()
                break
        else:
            raise StonerLoadError("Cannot find headers")
    skip = ix + i + 2
    column_headers = [f.strip() for f in line.strip().split("\t")]
    new_data.data = np.real(np.genfromtxt(new_data.filename, skip_header=skip, dtype=complex))
    new_data["dataset"] = dataset
    if "sld" in dataset.lower():
        new_data["type"] = "SLD"
    elif "asymmetry" in dataset.lower():
        new_data["type"] = "Asymmetry"
    elif "dd" in dataset.lower():
        new_data["type"] = "Down"
    elif "uu" in dataset.lower():
        new_data["type"] = "Up"
    new_data.column_headers = column_headers
    return new_data


@register_loader(
    patterns=(".ovf", 16),
    mime_types=[("text/plain", 16), ("application/octet-stream", 16)],
    name="OVFFile",
    what="Data",
)
def _load(new_data, filename=None, *args, **kargs):
    """Load function. File format has space delimited columns from row 3 onwards.

    Notes:
        This code can handle only the first segment in the data file.
    """
    if filename is None or not filename:
        new_data.get_filename("r")
    else:
        new_data.filename = filename
    # Reading in binary, converting to utf-8 where we should be in the text header
    with FileManager(new_data.filename, "rb") as data:
        line = data.readline().decode("utf-8", errors="ignore").strip("#\n \t\r")
        if line not in ["OOMMF: rectangular mesh v1.0", "OOMMF OVF 2.0"]:
            raise StonerLoadError("Not an OVF 1.0 or 2.0 file.")
        while _read_line(data, new_data.metadata):
            pass  # Read the file until we reach the start of the data block.
        if new_data.metadata["representation"] == "Binary":
            size = (  # Work out the size of the data block to read
                new_data.metadata["xnodes"]
                * new_data.metadata["ynodes"]
                * new_data.metadata["znodes"]
                * new_data.metadata["valuedim"]
                + 1
            ) * new_data.metadata["representation size"]
            bin_data = data.read(size)
            numbers = np.frombuffer(bin_data, dtype=f"<f{new_data.metadata['representation size']}")
            chk = numbers[0]
            if (
                chk != [1234567.0, 123456789012345.0][new_data.metadata["representation size"] // 4 - 1]
            ):  # If we have a good check number we can carry on, otherwise try the other endianness
                numbers = np.frombuffer(bin_data, dtype=f">f{new_data.metadata['representation size']}")
                chk = numbers[0]
                if chk != [1234567.0, 123456789012345.0][new_data.metadata["representation size"] // 4 - 1]:
                    raise StonerLoadError("Bad binary data for ovf gile.")

            data = np.reshape(
                numbers[1:],
                (new_data.metadata["xnodes"] * new_data.metadata["ynodes"] * new_data.metadata["znodes"], 3),
            )
        else:
            data = np.genfromtxt(
                data,
                max_rows=new_data.metadata["xnodes"] * new_data.metadata["ynodes"] * new_data.metadata["znodes"],
            )
    xmin, xmax, xstep = (
        new_data.metadata["xmin"],
        new_data.metadata["xmax"],
        new_data.metadata["xstepsize"],
    )
    ymin, ymax, ystep = (
        new_data.metadata["ymin"],
        new_data.metadata["ymax"],
        new_data.metadata["ystepsize"],
    )
    zmin, zmax, zstep = (
        new_data.metadata["zmin"],
        new_data.metadata["zmax"],
        new_data.metadata["zstepsize"],
    )
    Z, Y, X = np.meshgrid(
        np.arange(zmin + zstep / 2, zmax, zstep) * 1e9,
        np.arange(ymin + ystep / 2, ymax, ystep) * 1e9,
        np.arange(xmin + xstep / 2, xmax, xstep) * 1e9,
        indexing="ij",
    )
    new_data.data = np.column_stack((X.ravel(), Y.ravel(), Z.ravel(), data))

    column_headers = ["X (nm)", "Y (nm)", "Z (nm)", "U", "V", "W"]
    new_data.setas = "xyzuvw"
    new_data.column_headers = column_headers
    return new_data



# -*- coding: utf-8 -*-
"""Tests for Stoner.tools.formatting"""
import pytest
from Stoner.tools import formatting


def test_format_error():
    value = 1.2345e-6
    error = 2.456e-8
    assert formatting.format_error(value, error, fmt="text") == "0.00000123+/-0.00000002"
    assert formatting.format_error(value, error, fmt="html") == "0.00000123&plusmin;0.00000002"
    assert formatting.format_error(value, error, fmt="latex") == r"$0.00000123\pm 0.00000002$"

    assert formatting.format_error(value, error, fmt="text", mode="sci").strip() == "1.23+/-0.02E-6"
    assert (
        formatting.format_error(value, error, fmt="html", mode="sci").strip()
        == "1.23&plusmin;0.02&times; 10<sup>-6</sup>"
    )
    assert formatting.format_error(value, error, fmt="latex", mode="sci").strip() == r"$1.23\pm 0.02\times 10^{-6}\,$"

    assert formatting.format_error(value, error, fmt="text", mode="eng").strip() == "1.23+/-0.02 u"
    assert formatting.format_error(value, error, fmt="html", mode="eng").strip() == "1.23&plusmin;0.02 &micro;"
    assert formatting.format_error(value, error, fmt="latex", mode="sci").strip() == r"$1.23\pm 0.02\times 10^{-6}\,$"

    assert (
        formatting.format_error(value, error, fmt="latex", mode="eng", units="<bad\\hr>", escape=True).strip()
        == r"$1.23\pm 0.02\,\mathrm{\mu}\mathrm{\textlessbad\textbackslash{}hr\textgreater}$"
    )


def test_format_value():
    value = 1.2345e-6
    error = 0
    assert formatting.format_error(value, error, fmt="text") == "1.2345e-06"
    assert formatting.format_error(value, error, fmt="html") == "1.2345e-06"
    assert formatting.format_error(value, error, fmt="latex") == r"$1.2345e-06$"

    assert formatting.format_error(value, error, fmt="text", mode="sci").strip() == "1.2345000000000002E-6"
    assert (
        formatting.format_error(value, error, fmt="html", mode="sci").strip()
        == "1.2345000000000002&times; 10<sup>-6</sup>"
    )
    assert (
        formatting.format_error(value, error, fmt="latex", mode="sci").strip()
        == r"$1.2345000000000002\times 10^{-6}$"
    )

    assert formatting.format_error(value, error, fmt="text", mode="eng").strip() == "1.2345000000000002u"
    assert formatting.format_error(value, error, fmt="html", mode="eng").strip() == "1.2345000000000002&micro;"
    assert (
        formatting.format_error(value, error, fmt="latex", mode="sci").strip()
        == r"$1.2345000000000002\times 10^{-6}$"
    )

    assert (
        formatting.format_error(value, error, fmt="latex", mode="eng", units="<bad\\hr>", escape=True).strip()
        == r"$1.2345000000000002\mathrm{\mu}\mathrm{\textlessbad\textbackslash{}hr\textgreater}$"
    )


def test_ordinal():
    try:
        formatting.ordinal("G")
    except ValueError:
        pass
    else:
        assert False, "ordinal didn't raise a ValueError for non integer value"
    assert formatting.ordinal(11).endswith("th"), "Failed special handling for 11th in ordinal"
    assert formatting.ordinal(21).endswith("st"), "Failed to add st to 21st in ordinal"


if __name__ == "__main__":
    pytest.main(["--pdb", __file__])



# -*- coding: utf-8 -*-
"""Mxinin classes for DataFile objects."""

__all__ = ["DataFileSearchMixin"]

import copy
import numpy as np
from statsmodels.stats.weightstats import DescrStatsW

from ..compat import index_types, int_types
from ..tools import operator, isiterable, all_type
from ..tools.widgets import RangeSelect


class DataFileSearchMixin:
    """Mixin class that provides the search, selecting and sorting methods for a DataFile."""

    def _search_index(self, xcol=None, value=None, accuracy=0.0, invert=False):
        """Return an array of booleans for indexing matching rows for use with search method."""
        _ = self._col_args(scalar=False, xcol=xcol)
        x = self.find_col(_.xcol)
        if isinstance(value, (int_types, float)):  # Search for a value in x-column
            ix = np.isclose(self.data[:, x], value, atol=accuracy)
        elif isinstance(value, tuple) and len(value) == 2:  # within a range of values
            (low, u) = (min(value), max(value))
            low -= accuracy
            u += accuracy
            v = self.data[:, x]
            low = np.ones_like(v) * low
            u = np.ones_like(v) * u
            ix = np.logical_and(v > low, v <= u)
        elif (
            isinstance(value, (list, np.ndarray)) and all_type(value, bool) and len(value) <= len(self)
        ):  # index by list or array of booleans
            if len(value) < len(self):  # Expand array if necessary
                ix = np.append(value, [False] * len(self) - len(value))
            else:
                ix = np.array(value)
        elif isinstance(value, (list, np.ndarray)):  # Array or list of values
            ix = np.zeros(len(self), dtype=bool)
            for v in value:
                ix = np.logical_or(ix, self._search_index(xcol, v, accuracy))
        elif callable(value):  # Index with a callable function
            ix = np.array([value(r[x], r) for r in self], dtype=bool)
        elif value is None:  # If indexing with None, use a gui range slector
            selector = RangeSelect()
            ix = selector(self, x, accuracy)
        else:  # error!
            raise RuntimeError(f"Unknown search value type {value}")
        ix = np.logical_xor(invert, ix)
        if ix.ndim > 1:
            ix = ix[:, 0]
        return ix

    def asarray(self):
        """Provide a consistent way to get at the underlying array data."""
        return self.data

    def closest(self, value, xcol=None):
        """Return the row in a data file which has an x-column value closest to the given value.

        Args:
            value (float):
                Value to search for.

        Keyword Arguments:
            xcol (index or None):
                Column in which to look for value, or None to use setas.

        Returns:
            ndarray:
                A single row of data as a :py:class:`Stoner.Core.DataArray`.

        Notes:
            To find which row it is that has been returned, use the :py:attr:`Stoner.Core.DataArray.i`
            index attribute.
        """
        _ = self._col_args(xcol=xcol)
        xdata = np.abs(self // _.xcol - value)
        i = int(xdata.argmin())
        return self[i]

    def column(self, col):
        """Extract one or more columns of data from the datafile.

        Args:
            col (int, string, list or re):
                is the column index as defined for :py:meth:`DataFile.find_col`

        Returns:
            (ndarray):
                One or more columns of data as a :py:class:`numpy.ndarray`.
        """
        return self.data[:, self.find_col(col)]

    def find_col(self, col, force_list=False):
        """Indexes the column headers in order to locate a column of data.shape.

        Indexing can be by supplying an integer, a string, a regular expression, a slice or a list of any of the above.

        -   Integer indices are simply checked to ensure that they are in range
        -   String indices are first checked for an exact match against a column header
            if that fails they are then compiled to a regular expression and the first
            match to a column header is taken.
        -   A regular expression index is simply matched against the column headers and the
            first match found is taken. This allows additional regular expression options
            such as case insensitivity.
        -   A slice index is converted to a list of integers and processed as below
        -   A list index returns the results of feading each item in the list at :py:meth:`find_col`
            in turn.

        Args:
            col (int, a string, a re, a slice or a list):
                Which column(s) to retuirn indices for.

        Keyword Arguments:
            force_list (bool):
                Force the output always to be a list. Mainly for internal use only

        Returns:
            int, list of ints:
                The matching column index as an integer or a KeyError
        """
        return self.data._setas.find_col(col, force_list)

    def find_duplicates(self, xcol=None, delta=1e-8):
        """Find rows with duplicated values of the search column(s).

        Keyword Arguments:
            xcol (index types):
                The column)s) to search for duplicates in.
            delta (float or array):
                The absolute difference(s) to consider equal when comparing floats.

        Returns:
            (dictionary of value:[list of row indices]):
                The unique value and the associated rows that go with it.

        Notes:
            If *xcol* is not specified, then the :py:attr:`Data.setas` attribute is used. If this is also
            not set, then all columns are considered.
        """
        _ = self._col_args(xcol=xcol)
        if not _.has_xcol:
            _.xcol = list(range(self.shape[1]))
        search_data = self.data[:, _.xcol]
        if search_data.ndim == 1:
            search_data = np.atleast_2d(search_data).T

        delta = np.atleast_1d(np.array(delta))
        if delta.size != search_data.shape[1]:
            delta = np.append(delta, np.ones(search_data.shape[1]) * delta[0])[: search_data.shape[1]]
        results = dict()
        for ix in range(search_data.shape[0]):
            row = np.atleast_1d(search_data[ix])
            if tuple(row) in results:
                continue
            for iy, (value, dealt) in enumerate(zip(row, delta)):
                # Modify all search data that is close to the current row
                search_data[np.isclose(search_data[:, iy], value, atol=dealt), iy] = value
            matches = np.arange(search_data.shape[0])[np.all(search_data == row, axis=1)]
            results[tuple(row)] = matches.tolist()
        return results

    def remove_duplicates(self, xcol=None, delta=1e-8, strategy="keep first", ycol=None, yerr=None):
        """Find and remove rows with duplicated values of the search column(s).

        Keyword Arguments:
            xcol (index types):
                The column)s) to search for duplicates in.
            delta (float or array):
                The absolute difference(s) to consider equal when comparing floats.
            strategy (str, default *keep first*):
                What to do with duplicated rows. Options are:
                    - *keep first* - the first row is kept, others are discarded
                    - *average* - the duplicate rows are average together.
            ycol, yerr (index types):
                When using an average strategey identifies columns that represent values and uncertainties where
                the proper weighted standard error should be done.

        Returns:
            (dictionary of value:[list of row indices]):
                The unique value and the associated rows that go with it.

        Notes:
            If *ycol* is not specified, then the :py:attr:`Data.setas` attribute is used. If this is also
            not set, then all columns are considered.
        """
        _ = self._col_args(xcol=xcol, ycol=ycol, yerr=yerr, scalar=False)
        dups = self.find_duplicates(xcol=xcol, delta=delta)
        tmp = self.clone
        tmp.data = np.ma.empty((0, self.data.shape[1]))
        for indices in dups.values():
            section = self[indices, :]
            if strategy == "keep first":
                section = section[0, :]
            elif strategy == "average":
                tmp_sec = np.mean(section, axis=0)
                if _.has_ycol and _.has_yerr:  # reclaculate the ycolumns
                    ycol = _.ycol
                    yerr = _.yerr
                    if len(yerr) < len(ycol):
                        yerr += [yerr[0]] * (len(ycol) - len(yerr))
                    for yy, ye in zip(ycol, yerr):
                        stats = DescrStatsW(section[:, yy], weights=1 / (section[:, ye]) ** 2)
                        tmp_sec[yy] = stats.mean
                        tmp_sec[ye] = stats.std_mean
                section = tmp_sec
            else:
                raise RuntimeError(f"Unknown duplicate removal strategy {strategy}")
            tmp += section
        setas = self.setas
        self.data = tmp.data
        self.setas = setas
        return self

    def rolling_window(self, window=7, wrap=True, exclude_centre=False):
        """Iterate with a rolling window section of the data.

        Keyword Arguments:
            window (int):
                Size of the rolling window (must be odd and >= 3)
            wrap (bool):
                Whether to use data from the other end of the array when at one end or the other.
            exclude_centre (odd int or bool):
                Exclude the ciurrent row from the rolling window (defaults to False)

        Yields:
            ndarray:
                Yields with a section of data that is window rows long, each iteration moves the marker
                one row further on.
        """
        if isinstance(exclude_centre, bool) and exclude_centre:
            exclude_centre = 1
        if isinstance(exclude_centre, int_types) and not isinstance(exclude_centre, bool):
            if exclude_centre % 2 == 0:
                raise ValueError("If excluding the centre of the window, this must be an odd number of rows.")
            if window - exclude_centre < 2 or window < 3 or window % 2 == 0:
                raise ValueError(
                    """Window must be at least two bigger than the number of rows exluded from the centre, bigger than
                    3 and odd"""
                )

        hw = int((window - 1) / 2)
        if exclude_centre:
            hc = int((exclude_centre - 1) / 2)

        for i in range(len(self)):
            if i < hw:
                pre_data = self.data[i - hw :]
            else:
                pre_data = np.zeros((0, self.shape[1]))
            if i + 1 > len(self) - hw:
                post_data = self.data[0 : hw - (len(self) - i - 1)]
            else:
                post_data = np.zeros((0, self.shape[1]))
            starti = max(i - hw, 0)
            stopi = min(len(self), i + hw + 1)
            if exclude_centre:  # hacked to stop problems with DataArray concatenation
                tmp = self.clone  # copy all properties
                data = np.row_stack((self.data[starti : i - hc], self.data[i + 1 + hc : stopi]))
                tmp.data = np.array(data)  # guarantee an ndarray
                data = tmp.data  # get the DataArray
            else:
                data = self.data[starti:stopi]
            if wrap:
                tmp = self.clone  # copy all properties
                ret = np.row_stack((pre_data, data, post_data))
                tmp.data = np.array(ret)  # guarantee an ndarray
                ret = tmp.data  # get the DataArray
            else:
                ret = data
            yield ret

    def search(self, xcol=None, value=None, columns=None, accuracy=0.0):
        """Search the numerica data part of the file for lines that match and returns  the corresponding rows.

        Keyword Arguments:
            xcol (index types, None):
                a Search Column Index. If None (default), use the current setas.x
            value (float, tuple, list or callable, None):
                Value to look for
            columns (index or array of indices or None (default)):
                columns of data to return - none represents all columns.
            accuracy (float):
                Uncertainty to accept when testing equalities

        Returns:
            ndarray: numpy array of matching rows or column values depending on the arguments.

        Note:
            The value is interpreted as follows:

            - a float looks for an exact match
            - a list is a list of exact matches
            - an array or list of booleans (index like Numpy does)
            - a tuple should contain a (min,max) value.
            - A callable object should have accept a float and an array representing the value of
              the search col for the the current row and the entire row.
            - None opens an interactive span selector in a plot window.
        """
        ix = self._search_index(xcol, value, accuracy)
        if columns is None:  # Get the whole slice
            data = self.data[ix, :]
        else:
            columns = self.find_col(columns)
            if not isinstance(columns, list):
                data = self.data[ix, columns]
            else:
                data = self.data[ix, columns[0]]
                for c in columns[1:]:
                    data = np.column_stack((data, self.data[ix, c]))
        return data

    def section(self, **kargs):
        """Assuming data has x,y or x,y,z coordinates, return data from a section of the parameter space.

        Keyword Arguments:
            x (float, tuple, list or callable):
                x values ,atch this condition are included inth e section
            y (float, tuple, list  or callable):
                y values ,atch this condition are included inth e section
            z (float, tuple,list  or callable):
                z values ,atch this condition are included inth e section
            r (callable): a
            function that takes a tuple (x,y,z) and returns True if the line is to be included in section

        Returns:
            (DataFile):
                A :py:class:`DataFile` like object that includes only those lines from the original that match the
                section specification

        Internally this function is calling :py:meth:`DataFile.search` to pull out matching sections of the data array.
        To extract a 2D section of the parameter space orthogonal to one axis you just specify a condition on that
        axis. Specifying conditions on two axes will return a line of points along the third axis. The final
        keyword parameter allows you to select data points that lie in an arbitrary plane or line. eg::

            d.section(r=lambda x,y,z:abs(2+3*x-2*y)<0.1 and z==2)

        would extract points along the line 2y=3x+2 (note the use of an < operator to avoid floating point rounding
        errors) where the z-co-ordinate is 2.
        """
        cols = self.setas._get_cols()
        tmp = self.clone
        xcol = cols["xcol"] if cols.has_xcol else None
        ycol = cols["ycol"][0] if cols.has_ycol else None
        zcol = cols["zcol"][0] if cols.has_zcol else None

        accuracy = kargs.pop("accuracy", 0.0)

        if "x" in kargs:
            tmp.data = tmp.search(xcol, kargs.pop("x"), accuracy=accuracy)
        if "y" in kargs:
            tmp.data = tmp.search(ycol, kargs.pop("y"), accuracy=accuracy)
        if "z" in kargs:
            tmp.data = tmp.search(zcol, kargs.pop("z"), accuracy=accuracy)
        if "r" in kargs:
            func = lambda x, r: kargs.pop("r")(r[xcol], r[ycol], r[zcol])
            tmp.data = tmp.search(0, func, accuracy=accuracy)

        if kargs:  # Fallback to working with select if nothing else.
            tmp.select(**kargs)
        return tmp

    def select(self, *args, **kargs):
        """Produce a copy of the DataFile with only data rows that match a criteria.

        Args:
            args (various):
                A single positional argument if present is interpreted as follows:

                -   If a callable function is given, the entire row is presented to it. If it evaluates True then that
                    row is selected. This allows arbitrary select operations
                -   If a dict is given, then it and the kargs dictionary are merged and used to select the rows

        Keyword Arguments:
            kargs (various):
                Arbitrary keyword arguments are interpreted as requestion matches against the corresponding
                columns. The keyword argument may have an additional *__operator** appended to it which is interpreted
                as follows:

                -   *eq*  value equals argument value (this is the default test for scalar argument)
                -   *ne*  value doe not equal argument value
                -   *gt*  value doe greater than argument value
                -   *lt*  value doe less than argument value
                -   *ge*  value doe greater than or equal to argument value
                -   *le*  value doe less than or equal to argument value
                -   *between*  value lies between the minimum and maximum values of the argument (the default test
                    for 2-length tuple arguments)
                -   *ibetween*,*ilbetween*,*iubetween* as above but include both,lower or upper values

        Returns:
            (DatFile): a copy the DataFile instance that contains just the matching rows.

        Note:
            if the operator is preceeded by *__not__* then the sense of the test is negated.

            If any of the tests is True, then the row will be selected, so the effect is a logical OR. To
            achieve a logical AND, you can chain two selects together::

                d.select(temp__le=4.2,vti_temp__lt=4.2).select(field_gt=3.0)

            will select rows that have either temp or vti_temp metadata values below 4.2 AND field metadata values
            greater than 3.

            If you need to select on a row value that ends in an operator word, then append
            *__eq* in the keyword name to force the equality test. If the metadata keys to select on are not valid
            python identifiers, then pass them via the first positional dictionary value.

            There is a "magic" column name "_i" which is interpreted as the row numbers of the data.

        Example
            .. plot:: samples/select_example.py
                :include-source:
                :outname: select
        """
        if len(args) == 1:
            if callable(args[0]):
                kargs["__"] = args[0]
            elif isinstance(args[0], dict):
                kargs.update(args[0])
        result = self.clone
        res = np.zeros(len(self), dtype=bool)
        for arg in kargs:
            parts = arg.split("__")
            if parts == ["", ""]:
                func = kargs[arg]
                res = np.logical_or(res, np.array([func(r) for r in self.data]))
                continue
            if len(parts) == 1 or parts[-1] not in operator:
                parts.append("eq")
            if len(parts) > 2 and parts[-2] == "not":
                end = -2
                negate = True
            else:
                end = -1
                negate = False
            if parts[0] == "_i":
                res = np.logical_or(res, np.logical_xor(negate, operator[parts[-1]](self.data.i, kargs[arg])))
            else:
                col = "__".join(parts[:end])
                res = np.logical_or(res, np.logical_xor(negate, operator[parts[-1]](self.column(col), kargs[arg])))
        result.data = self.data[res, :]
        return result

    def sort(self, *order, **kargs):
        """Sort the data by column name.

        Arguments:
            order (column index or list of indices or callable function):
                One or more sort order keys.

        Keyword Arguments:
            reverse (boolean):
                If true, the sorted array isreversed.

        Returns:
            (self):
                A copy of the :py:class:`DataFile` sorted object

        Notes:
            Sorts in place and returns a copy of the sorted data object fo chaining methods.

            If the argument is a callable function then it should take a two tuple arguments and
            return +1,0,-1 depending on whether the first argument is bigger, equal or smaller. Otherwise
            if the argument is interpreted as a column index. If a single argument is supplied, then it may be
            a list of column indices. If no sort orders are supplied then the data is sorted by the
            :py:attr:`DataFile.setas` attribute or if that is not set, then order of the columns in the data.
        """
        reverse = kargs.pop("reverse", False)
        order = list(order)
        setas = self.setas.clone
        ch = copy.copy(self.column_headers)
        if not order:
            if self.setas.cols["xcol"] is not None:
                order = [self.setas.cols["xcol"]]
            order.extend(self.setas.cols["ycol"])
            order.extend(self.setas.cols["zcol"])
        if not order:  # Ok, no setas here then
            order = None
        elif len(order) == 1:
            order = order[0]

        if order is None:
            order = list(range(len(self.column_headers)))
        recs = self.records
        if callable(order):
            d = sorted(recs, cmp=order)
        elif isinstance(order, index_types):
            order = [recs.dtype.names[self.find_col(order)]]
            d = np.sort(recs, order=order)
        elif isiterable(order):
            order = [recs.dtype.names[self.find_col(x)] for x in order]
            d = np.sort(recs, order=order)
        else:
            raise KeyError(f"Unable to work out how to sort by a {type(order)}")
        self.data = d.view(dtype=self.dtype).reshape(len(self), len(self.column_headers))
        if reverse:
            self.data = self.data[::-1]
        self.data._setas = setas
        self.column_headers = ch
        return self

    def split(self, *args, final="files"):
        """Recursively splits the current DataFile into a :py:class:`Stoner.Folders.DataFolder`.

        Args:
            *args (column index or function):
                Each argument is used in turn to find key values for the files in the DataFolder

        Keyword Arguments:
            final (str):
                Controls whether the final argument plaes the files in the DataFolder (default: "files") or in
                groups ("groups")

        Returns:
            Stoner.Folders.DataFolder:
                A :py:class:`Stoner.Folders.DataFolder` object containing the individual
                :py:class:`AnalysisMixin` objects

        Note:
            Creates a DataFolder of  DataFiles where each one contains the rows from the original object which
            had the same value of a given column(s) or function.


            On each iteration the first argument is called. If it is a column type then rows which amtch each unique
            value are collated together and made into a separate file. If the argument is a callable, then it is
            called for each row, passing the row as a single 1D array and the return result is used to group lines
            together. The return value should be hashable.

            Once this is done and the :py:class:`Stoner.Folders.DataFolder` exists, if there are remaining argument,
            then the method is called recusivelyt for each file and the resulting DataFolder added into the root
            DataFolder and the file is removed.

            Thus, when all of the arguments are evaluated, the resulting DataFolder is a multi-level tree.

            .. warning::

                There has been a change in the arguments for the split function  from version 0.8 of the Stoner
                Package.
        """
        from Stoner import DataFolder

        if not args:
            xcol = self.setas._get_cols("xcol")
        else:
            args = list(args)
            xcol = args.pop(0)
        data = dict()

        if isinstance(xcol, index_types):
            for val in np.unique(self.column(xcol)):
                newfile = self.clone
                newfile.filename = f"{self.column_headers[self.find_col(xcol)]}={val} {self.filename}"
                newfile.data = self.search(xcol, val)
                data[val] = newfile
        elif callable(xcol):
            try:  # Try to call function with all data in one go
                keys = xcol(self.data)
                if not isiterable(keys):
                    keys = [keys] * len(self)
            except Exception:  # pylint: disable=W0703  # Ok try instead to do it row by row
                keys = [xcol(r) for r in self]
            if not isiterable(keys) or len(keys) != len(self):
                raise RuntimeError("Not returning an index of keys")
            keys = np.array(keys)
            for key in np.unique(keys):
                data[key] = self.clone
                data[key].data = self.data[keys == key, :]
                data[key].filename = f"{xcol.__name__}={key} {self.filename}"
                data[key].setas = self.setas
        else:
            raise NotImplementedError(f"Unable to split a file with an argument of type {type(xcol)}")
        out = DataFolder(nolist=True, setas=self.setas)
        for k, f in data.items():
            if args:
                out.add_group(k)
                out.groups[k] = f.split(*args)
            else:
                if final == "files":
                    out += f
                elif final == "groups":
                    out.add_group(k)
                    f.filename = self.filename
                    out.groups[k] += f
                else:
                    raise ValueError(f"{final} not recognised as a valid value for final")
        return out

    def unique(self, col, return_index=False, return_inverse=False):
        """Return the unique values from the specified column - pass through for numpy.unique.

        Args:
            col (index):
                Column to look for unique values in

        Keyword Arguments:
            return_index (bool):
                Pass through to :py:func:`np.unique`
            reverse (bool):
                Pass through to :py:func:`np.unique`

        Returns:
            (1D array):
                Array of unique values from the column.
        """
        return np.unique(self.column(col), return_index, return_inverse)



# -*- coding: utf-8 -*-
"""Kerr Image Processing Module.

Created on Fri Apr 21 17:29:08 2017
Derivatives of ImageArray and ImageStack specific to processing Kerr images.

@author: phyrct
"""
__all__ = ["KerrArray", "KerrStack", "MaskStack"]

import os

import numpy as np

from ..Image import ImageArray, ImageStack, ImageFile
from ..tools import make_Data
from ..tools.decorators import class_modifier, image_file_adaptor
from . import kerrfuncs

try:
    import pytesseract  # pylint: disable=unused-import

    _tesseractable = True
except ImportError:
    pytesseract = None
    _tesseractable = False

GRAY_RANGE = (0, 65535)  # 2^16
IM_SIZE = (512, 672)  # Standard Kerr image size
AN_IM_SIZE = (554, 672)  # Kerr image with annotation not cropped
pattern_file = os.path.join(os.path.dirname(__file__), "kerr_patterns.txt")


@class_modifier(kerrfuncs)
class KerrArray(ImageArray):
    """A subclass for Kerr microscopy specific image functions."""

    # useful_keys are metadata keys that we'd usually like to keep from a
    # standard kerr output.

    def __init__(self, *args, **kargs):
        """Initialise KerrArray as a subclasses ImageArray.

        Extra keyword arguments accepted are given below.

        Keyword Arguments:
            reduce_metadata(bool):
                if True reduce the metadata to useful bits and do some processing on it
            asfloat(bool)
                if True convert the image to float values between 0 and 1 (necessary
                for some forms of processing)
            crop_text(bool):
                whether to crop the bottom text area from the image
            ocr_metadata(bool):
                whether to try to use optical character recognition to get the
                metadata from the image (necessary for images taken pre 06/2016
                and so far field from hysteresis images)
            field_only(bool):
                if ocr_metadata is true, get field only (bit faster)
        """
        kerrdefaults = {
            "ocr_metadata": False,
            "field_only": False,
            "reduce_metadata": True,
            "asfloat": True,
            "crop_text": True,
        }
        kerrdefaults.update(kargs)
        super().__init__(*args, **kargs)
        self._tesseractable = None
        if kerrdefaults["reduce_metadata"]:
            self.reduce_metadata()
        if kerrdefaults["ocr_metadata"]:
            self.ocr_metadata(field_only=kerrdefaults["field_only"])
        if kerrdefaults["asfloat"]:
            self.asfloat()
        if kerrdefaults["crop_text"]:
            self.crop_text()

    @property
    def tesseractable(self):
        """Do a test call to tesseract to see if it is there and cache the result."""
        return _tesseractable

    def save(self, filename=None, **kargs):
        """Stub method for a save function."""
        raise NotImplementedError(f"Save is not implemented in {self.__class__}")


@class_modifier(kerrfuncs, adaptor=image_file_adaptor)
class KerrImageFile(ImageFile):
    """Subclass of ImageFile that keeps the data as a KerrArray so that extra functions are available."""

    priority = 16
    mime_type = ["image/png"]
    pattern = ["*.png"]

    def __init__(self, *args, **kargs):
        """Ensure that the image is a KerrImage."""
        super().__init__(*args, **kargs)
        self._image = self.image.view(KerrArray)

    @ImageFile.image.getter
    def image(self):  # pylint disable=invalid-overridden-method
        """Access the image data."""
        return self._image.view(KerrArray)

    @ImageFile.image.setter
    def image(self, v):  # pylint: disable=function-redefined
        """Ensure stored image is always an ImageArray."""
        filename = self.filename
        v = KerrArray(v)
        # ensure setting image goes into the same memory block if from stack
        if (
            hasattr(self, "_fromstack")
            and self._fromstack
            and self._image.shape == v.shape
            and self._image.dtype == v.dtype
        ):
            self._image[:] = v
            self._image = self._image.view(KerrArray)
        else:
            self._image = KerrArray(v)
        self.filename = filename


class KerrStackMixin:
    """A mixin for :py:class:`ImageStack` that adds some functionality particular to Kerr images.

    Attributes:
        fields(list):
            list of applied fields in stack. This is the most important metadata
            for things like hysteresis.
    """

    _defaults = {"type": KerrImageFile}

    @property
    def fields(self):
        """Produce an array of field values from the metadata."""
        if not hasattr(self, "_field"):
            if "field" not in self.metadata:
                self._field = np.arange(len(self))
            else:
                self._field = np.array(self.metadata["field"])
        return self._field

    def crop_text(self):
        """Crop the bottom text area from a standard Kermit image across the complete stack.

        Returns:
        (ImageArray):
            cropped image
        """
        images = self.shape[0]
        if self.shape[1:3] == IM_SIZE:
            return self
        if self.shape[1:3] != AN_IM_SIZE:
            raise ValueError(
                f"Need a full sized Kerr image to crop. Current size is {self.shape}"
            )  # check it's a normal image
        self._sizes = np.column_stack(
            (np.ones(images, dtype=int) * IM_SIZE[0], np.ones(images, dtype=int) * IM_SIZE[1])
        )
        new_size = self.max_size + (images,)
        self._resize_stack(new_size)
        return self

    def hysteresis(self, mask=None):
        """Make a hysteresis loop of the average intensity in the given images.

        Keyword Argument:
            mask(ndarray or list):
                boolean array of same size as an image or imarray or list of
                masks for each image. If True then don't include that area in
                the intensity averaging.

        Returns
        -------
        hyst(Data):
            'Field', 'Intensity', 2 column array
        """
        hyst = np.column_stack((self.fields, np.zeros(len(self))))
        for i, im in enumerate(self):
            if isinstance(im, ImageFile):
                im = im.image
            if isinstance(mask, np.ndarray) and mask.ndim == 2:
                hyst[i, 1] = np.average(im[np.invert(mask.astype(bool))])
            elif isinstance(mask, np.ndarray) and mask.ndim == 3:
                hyst[i, 1] = np.average(im[np.invert(mask[i, :, :].astype(bool))])
            elif isinstance(mask, (tuple, list)):
                hyst[i, 1] = np.average(im[np.invert(mask[i])])
            else:
                hyst[i, 1] = np.average(im)
        d = make_Data(hyst, setas="xy")
        d.column_headers = ["Field", "Intensity"]
        return d

    def index_to_field(self, index_map):
        """Convert an image of index values into an image of field values."""
        fieldvals = np.take(self.fields, index_map)
        return ImageArray(fieldvals)

    def denoise_thresh(self, denoise_weight=0.1, thresh=0.5, invert=False):
        """Apply denoise then threshold images.

        Return:
            (ndarray) MaskStack:
                True for values greater than thresh, False otherwise
                else return True for values between thresh and 1
        """
        masks = self.clone
        masks.each.denoise(weight=denoise_weight)
        masks.each.threshold_minmax(threshmin=thresh, threshmax=np.max(masks.imarray))
        masks = MaskStack(masks)
        if invert:
            masks.stack = ~masks.stack  # pylint: disable=attribute-defined-outside-init
        return masks

    def find_threshold(self, testim=None, mask=None):
        """Try to find the threshold value at which the image switches.

        Takes it as the median value of the testim. Masks values
        where the difference is less than tolerance in case part of the image is
        irrelevant.
        """
        if testim is None:
            testim = self[len(self) // 2]
        elif isinstance(testim, (int, str)):
            testim = self[testim]
        elif isinstance(testim, np.ndarray) and testim.shape == self[len(self) // 2].shape:
            pass
        else:
            raise ValueError("Cannot find testimage for thresholding.")
        if mask is None:
            med = testim.median()
        else:
            med = testim[~testim.mask]
        return med

    def stable_mask(self, tolerance=1e-2, comparison=(0, -1)):
        """Produce a mask of areas of the image that are changing little over the stack.

        comparison is an optional tuple that gives the index of two images
        to compare, otherwise first and last used. tolerance is the difference
        tolerance
        """
        first, last = comparison
        mask = np.zeros(self[0].shape, dtype=bool)
        mask[abs(self[last] - self[first]) < tolerance] = True
        return mask

    def HcMap(
        self,
        threshold=0.5,
        correct_drift=False,
        baseimage=0,
        quiet=True,
        saturation_end=True,
        saturation_white=True,
        extra_info=False,
    ):
        """Produce a map of the switching field at every pixel in the stack.

        It needs the stack to start saturated one way and end saturated the other way.

        Keyword Arguments:
            threshold(float):
                the threshold value for the intensity switching. This will need to
                be tuned for each stack
            correct_drift(bol):
                whether to correct drift on the image stack before proceding
            baseimage(int or ImageArray):
                we use drift correction from the baseimage.
            saturation_end(bool):
                last image in stack is closest to saturation
            saturation_white(bool):
                bright pixels are saturated dark pixels are not yet switched
            quiet: bool
                choose wether to output status updates as print messages
            extra_info(bool):
                choose whether to return intermediate calculation steps as an extra dictionary
        Returns:
            (ImageArray): The map of field values for switching of each pixel in the stack
        """
        ks = self.clone
        if isinstance(baseimage, int):
            baseimage = self[baseimage].clone
        elif isinstance(baseimage, np.ndarray):
            baseimage = baseimage.view(ImageArray)
        if correct_drift:
            ks.apply_all("correct_drift", ref=baseimage, quiet=quiet)
            if not quiet:
                print("drift correct done")
        masks = self.denoise_thresh(denoise_weight=0.1, thresh=threshold, invert=not (saturation_white))
        if not quiet:
            print("thresholding done")
        si, sp = masks.switch_index(saturation_end=saturation_end)
        Hcmap = ks.index_to_field(si)
        Hcmap[Hcmap == ks.fields[0]] = 0  # not switching does not give us a Hc value
        if extra_info:
            ei = {"switch_index": si, "switch_array": sp, "masks": masks}
            return Hcmap, ei
        return Hcmap

    def average_Hcmap(self, weights=None, ignore_zeros=False):
        """Get an array of average pixel values for the stack.

        Return average of pixel values in the stack.

        Keyword Arguments:
            ignore zeros(bool):
                Weight zero values in an image as 0 in the averaging.

        Returns:
            average(ImageArray):
                average values
        """
        if ignore_zeros:
            weights = self.clone
            weights.imarray = weights.imarray.astype(bool).astype(int)  # 1 if Hc isn't zero, zero otherwise
            condition = np.sum(weights, axis=0) == 0  # stop zero division error
            for m in range(self.shape[0]):
                weights[m] = np.select([condition, np.logical_not(condition)], [np.ones_like(weights[m]), weights[m]])
            # weights means we only account for non-zero values in average
        average = np.average(self.imarray, axis=0, weights=weights)
        return average.view(ImageArray)


class MaskStackMixin:
    """A Mixin for :py:class:`Stoner.Image.ImageStack` but made for stacks of boolean or binary images."""

    def __init__(self, *args, **kargs):
        """Ensure the data is boolean."""
        super().__init__(*args, **kargs)
        self._stack = self._stack.astype(bool)

    def switch_index(self, saturation_end=True, saturation_value=True):
        """Construct a map of switching points in a hysteresis stack.

        Given a stack of boolean masks representing a hystersis loop find the stack index of the saturation
        field for each pixel.

        Take the final mask as all switched (or the first mask if saturation_end
        is False). Work back through the masks taking the first time a pixel
        switches as its coercive field (ie the last time it switches before
        reaching saturation).
        Elements that start switched at the lowest measured field or never
        switch are given a zero index.

        At the moment it's set up to expect masks to be false when the sample is saturated
        at a high field

        Keyword Arguments:
            saturation_end(bool):
                True if the last image is closest to the fully saturated state.
                False if you want the first image
            saturation_value(bool):
                if True then a pixel value True means that switching has occurred
                (ie magnetic saturation would be all True)

        Returns:
            switch_ind: MxN ndarray of int
                index that each pixel switches at
            switch_progession: MxNx(P-1) ndarray of bool
                stack of masks showing when each pixel saturates

        """
        ms = self.clone
        if not saturation_end:
            ms = ms.reverse()
        # arr1 = ms[0].astype(float) #find out whether True is at begin or end
        # arr2 = ms[-1].astype(float)
        # if np.average(arr1)>np.average(arr2): #OK so it's bright at the start
        if not saturation_value:
            self.imarray = np.invert(ms.imarray)  # Now it's bright (True) at end
        switch_ind = np.zeros(ms[0].shape, dtype=int)
        switch_prog = self.clone
        switch_prog.imarray = np.zeros(self.shape, dtype=bool)
        del switch_prog[-1]
        for m in reversed(range(len(ms) - 1)):  # go from saturation backwards
            already_done = np.copy(switch_ind).astype(dtype=bool)  # only change switch_ind if it hasn't already
            condition = np.logical_and(not ms[m], ms[m + 1])
            condition = np.logical_and(condition, np.invert(already_done))
            condition = [condition, np.logical_not(condition)]
            choice = [np.ones(switch_ind.shape) * m, switch_ind]  # index or leave as is
            switch_ind = np.select(condition, choice)
            switch_prog[m] = already_done
        if not saturation_end:
            switch_ind = -switch_ind + len(self) - 1  # should check this!
            switch_prog.reverse()
        switch_ind = ImageArray(switch_ind.astype(int))
        return switch_ind, switch_prog


class KerrStack(KerrStackMixin, ImageStack):
    """Represent a stack of Kerr images."""


class MaskStack(MaskStackMixin, KerrStackMixin, ImageStack):
    """Represent a set of masks for Kerr images."""



# -*- coding: utf-8 -*-
"""
test_Core.py
Created on Tue Jan 07 22:05:55 2014

@author: phygbu
"""

import pytest
import sys
import os.path as path
import numpy as np

pth = path.dirname(__file__)
pth = path.realpath(path.join(pth, "../../"))
sys.path.insert(0, pth)
from Stoner import Data


"""Path to sample Data File"""
slfdatadir = path.join(pth, "sample-data")

np.random.seed(12345)
slfd1 = Data(path.join(slfdatadir, "OVF1.ovf"))
slfd2 = Data(path.join(slfdatadir, "TDI_Format_RT.txt"))
slfd3 = Data(path.join(slfdatadir, "New-XRay-Data.dql"))
slfd4 = Data(np.column_stack([np.ones(100), np.ones(100) * 2]), setas="xy")


def test_functions():
    # Test section:
    _ = slfd1.section(z=(12, 13))
    f = slfd2.split(lambda r: r["Temp"] < 150)
    assert len(f[0]) == 838, "Split failed to work."
    assert len(slfd3.threshold(2000, rising=True, falling=True, all_vals=True)) == 5, "Threshold failure."


def test_peaks():
    d = slfd3.clone
    d.peaks(width=8, poly=4, significance=100, modify=True)
    assert len(d) == 11, "Failed on peaks test."


def test_threshold():
    # set up some zigzag data
    # mins at 0,100,200,300,400, max at 50, 150, 250, 350 and zeros in between
    ar = np.zeros((400, 2))
    ar[:, 0] = np.arange(0, len(ar))
    for i in range(4):
        ar[i * 100 : i * 100 + 50, 1] = np.linspace(-1, 1, 50)
    for i in range(4):
        ar[i * 100 + 50 : i * 100 + 100, 1] = np.linspace(1, -1, 50)
    d = Data(ar, setas="xy")
    assert len(d.threshold(0, rising=True, falling=False, all_vals=True) == 4)
    assert len(d.threshold(0, rising=False, falling=True, all_vals=True) == 4)
    assert len(d.threshold(0, interpolate=False, rising=False, falling=True, all_vals=True) == 4)
    assert d.threshold(0, all_vals=True)[1] == 124.5
    _ = d
    assert (
        np.sum(d.threshold([0.0, 0.5, 1.0]) - np.array([[24.5, 36.74999999, 49.0]])) < 1e-6
    ), "Multiple threshold failed."
    assert np.isclose(
        d.threshold(0, interpolate=False, all_vals=True)[1], 124.5, 6
    ), "Threshold without interpolation failed."
    result = d.threshold(0, interpolate=False, all_vals=True, xcol=False)
    assert np.allclose(
        result, np.array([[24.5, 0.0], [124.5, 0.0], [224.5, 0.0], [324.5, 0.0]])
    ), "Failed threshold with False scol - result was {}".format(result)


def test_apply():
    slfapp = Data(np.zeros((100, 1)), setas="y")
    slfapp.apply(lambda r: r.i[0], header="Counter")

    def calc(r, omega=1.0, k=1.0):
        return np.sin(r.y * omega)

    slfapp.apply(calc, replace=False, header="Sin", _extra={"omega": 0.1}, k=1.0)
    slfapp.apply(lambda r: r.__class__([r[1], r[0]]), replace=True, header=["Index", "Sin"])
    slfapp.setas = "xy"
    assert np.isclose(slfapp.integrate(output="result"), 18.87616564214), "Integrate after applies failed."


def test_scale():
    x = np.linspace(-5, 5, 101)
    y = np.sin(x)
    orig = Data(x + np.random.normal(size=101, scale=0.025), y + np.random.normal(size=101, scale=0.01))
    orig.setas = "xy"

    XTests = [[(0, 0, 0.5), (0, 2, -0.1)], [(0, 0, 0.5)], [(0, 2, -0.2)]]
    YTests = [[(1, 1, 0.5), (1, 2, -0.1)], [(1, 1, 0.5)], [(1, 2, -0.2)]]
    for xmode, xdata, xtests in zip(["linear", "scale", "offset"], [x * 2 + 0.2, x * 2, x + 0.2], XTests):
        for ymode, ydata, ytests in zip(["linear", "scale", "offset"], [y * 2 + 0.2, y * 2, y + 0.2], YTests):
            to_scale = Data(
                xdata + np.random.normal(size=101, scale=0.025), ydata + np.random.normal(size=101, scale=0.01)
            )
            to_scale.setas = "xy"
            to_scale.scale(orig, xmode=xmode, ymode=ymode)
            transform = to_scale["Transform"]
            t_err = to_scale["Transform Err"]
            for i, j, v in xtests + ytests:
                assert (
                    np.abs(transform[i, j] - v) <= 5 * t_err[i, j]
                ), "Failed to get correct trandorm factor for {}:{} ({} vs {}".format(
                    xmode, ymode, transform[i, j], v
                )

    to_scale = Data(
        x * 5 + 0.1 + np.random.normal(size=101, scale=0.025),
        y * 0.5 + 0.1 + 0.5 * x + np.random.normal(size=101, scale=0.01),
    )
    to_scale.setas = "xy"
    to_scale.scale(orig, xmode="affine")
    a_tranform = np.array([[0.2, 0.0, -0.02], [-0.2, 2.0, -0.17]])
    t_delta = np.abs(to_scale["Transform"] - a_tranform)
    t_in_range = t_delta < to_scale["Transform Err"] * 5
    assert np.all(t_in_range), "Failed to produce correct affine scaling {} vs {}".format(
        to_scale["Transform"], a_tranform
    )


def test_clip():
    x = np.linspace(0, np.pi * 10, 1001)
    y = np.sin(x)
    z = np.cos(x)
    d = Data(x, y, z, setas="xyz")
    d.clip((-0.1, 0.2), "Column 2")
    assert (d.z.min() >= -0.1) and (d.z.max() <= 0.2), "Clip with a column specified failed."
    d = Data(x, y, z, setas="xyz")
    d.clip((-0.5, 0.7))
    assert (d.y.min() >= -0.5) and (d.y.max() <= 0.7), "Clip with no column specified failed."


def test_integrate():
    d = Data(path.join(slfdatadir, "SLD_200919.dat"))
    d.setas = "x..y"
    d.integrate(result=True, header="Total_M")
    result = d["Total_M"]
    assert np.isclose(result, 4.19687459365, 7), "Integrate returned the wrong result!"
    d.setas[-1] = "y"
    d.plot(multiple="y2")
    assert len(d.axes) == 2, "Failed to produce plot with double y-axis"
    d.close("all")
    d.setas = "x..y"
    fx = d.interpolate(None)
    assert fx(np.linspace(1, 1500, 101)).shape == (101, 7), "Failed to get the interpolated shape right"


def test_sg_filter():
    x = np.linspace(0, 10 * np.pi, 1001)
    y = np.sin(x) + np.random.normal(size=1001, scale=0.05)
    d = Data(x, y, column_headers=["Time", "Signal"], setas="xy")
    d.SG_Filter(order=1, result=True)
    d.setas = "x.y"
    d.y = d.y - np.cos(x)
    assert np.isclose(d.y[5:-5].mean(), 0, atol=0.01), "Failed to differentiate correctly"


if __name__ == "__main__":  # Run some tests manually to allow debugging
    pytest.main(["--pdb", __file__])



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""Fit Ic(B) to Airy function."""
# pylint: disable=invalid-name
import os

from Stoner import Data, __home__
from Stoner.analysis.fitting.models.superconductivity import Ic_B_Airy

os.chdir(os.path.join(__home__, "..", "doc", "samples", "Fitting"))

data = Data("data/Ic_B.txt", setas={"x": "Magnet Output", "y": "Ic"})

data.lmfit(Ic_B_Airy, result=True, header="Fit")

data.setas = {"x": "Magnet Output", "y": ["Ic", "Fit"]}
data.plot(fmt=["r+", "b-"])

data.annotate_fit(Ic_B_Airy, mode="eng", x=0.6, y=0.5, fontsize="small")

data.title = r"Critical current vs Field for $4\mu m$ junction"
data.xlabel = r"Magnetic Field $\mu_0H (\mathrm{T})$"
data.ylabel = r"Critical Current $I_c (A)$"



"""Example plot style using Seaborn plot styling template."""

# pylint: disable=invalid-name
import os.path as path

from Stoner import Data, __home__
from Stoner.plot.formats import SeabornPlotStyle

filename = path.realpath(
    path.join(__home__, "..", "doc", "samples", "sample.txt")
)
d = Data(
    filename,
    setas="xyy",
    template=SeabornPlotStyle(
        stylename="dark", context="talk", palette="muted"
    ),
)
d.plot(multiple="y2")



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""Implement DataFile classes for some generic file formats."""
import contextlib
import io
import re
import sys
import logging

from ...core.exceptions import StonerLoadError
from ...Image import ImageArray
from ...formats.decorators import register_loader


class _refuse_log(logging.Filter):
    """Refuse to log all records."""

    def filter(self, record):
        """Do not log anything."""
        return False


@contextlib.contextmanager
def catch_sysout(*args):
    """Temporarily redirect sys.stdout and.sys.stdin."""
    stdout, stderr = sys.stdout, sys.stderr
    out = io.StringIO()
    sys.stdout, sys.stderr = out, out
    logger = logging.getLogger("hyperspy.io")
    logger.addFilter(_refuse_log)
    yield None
    logger.removeFilter(_refuse_log)
    sys.stdout, sys.stderr = stdout, stderr
    return


def _delim_detect(line):
    """Detect a delimiter in a line.

    Args:
        line(str):
            String to search for delimiters in.

    Returns:
        (str):
            Delimiter to use.

    Raises:
        StnerLoadError:
            If delimiter cannot be located.
    """
    quotes = re.compile(r"([\"\'])[^\1]*\1")
    line = quotes.sub("", line)  # Remove quoted strings first
    current = (None, len(line))
    for delim in "\t ,;":
        try:
            idx = line.index(delim)
        except ValueError:
            continue
        if idx < current[1]:
            current = (delim, idx)
    if current[0] is None:
        raise StonerLoadError("Unable to find a delimiter in the line")
    return current[0]


@register_loader(
    patterns=[(".tif", 8), (".tiff", 8), (".png", 8), (".npy", 8)],
    mime_types=[("image/tiff", 8), ("image/png", 8), ("application/octet-stream", 8)],
    name="ImageFile",
    what="Image",
)
def load_imagefile(new_image, filename, *args, **kargs):
    """Load an ImageFile by calling the ImageArray method instead."""
    new_image._image = ImageArray(filename, *args, **kargs)
    for k in new_image._image._public_attrs:
        setattr(new_image, k, getattr(new_image._image, k, None))
    return new_image
